from datetime import datetime
import pandas as pd
import pytz
from time import time

from typing import Tuple

# from .quest import get_server_time
from .logs import Logs

time_logger = Logs(filename='./logs/time_helpers.log', name='time_helpers', info_level='INFO')

tick_seconds = {'1m': 60, '3m': 60 * 3, '5m': 5 * 60, '15m': 15 * 60, '30m': 30 * 60, '1h': 60 * 60, '2h': 60 * 60 * 2,
                '4h': 60 * 60 * 4, '6h': 60 * 60 * 6, '8h': 60 * 60 * 8, '12h': 60 * 60 * 12, '1d': 60 * 60 * 24,
                '3d': 60 * 60 * 24 * 3, '1w': 60 * 60 * 24 * 7, '1M': 60 * 60 * 24 * 30}
pandas_freq_tick_interval = {'1m': '1T',
                             '3m': '3T',
                             '5m': '5T',
                             '15m': '15T',
                             '30m': '30T',
                             '1h': '1H',
                             '2h': '2H',
                             '4h': '4H',
                             '6h': '6H',
                             '8h': '8H',
                             '12h': '12H',
                             '1d': '1D',
                             '3d': '3D',
                             '1w': '1W',
                             '1M': '1M'}

tick_interval_values = ['1m', '3m', '5m', '15m', '30m', '1h', '2h', '4h', '6h', '8h', '12h', '1d', '3d', '1w', '1M']


# time control functions

def convert_utc_ms_column_to_time_zone(df: pd.DataFrame, col: str, time_zone='Europe/Madrid') -> pd.Series:
    """
    Replace a column from milliseconds to datetime with time zone.
    :param df: A pandas dataframe.
    :param col: Name of the column to replace.
    :param time_zone: A time zone like 'Europe/Madrid'
    :return: Modified dataframe.
    """
    df[col] = pd.to_datetime(df[col], unit='ms')
    return df[col].dt.tz_localize('utc').dt.tz_convert(time_zone)


def convert_datetime_to_string(dt: datetime) -> str:
    """Convierte datetime tanto utc como local. No cambia zona horaria"""
    return dt.strftime("%Y-%m-%d %H:%M:%S")


def convert_datetime_to_milliseconds(dt: datetime, timezoned: str = None) -> float:
    if not timezoned:
        epoch = datetime.utcfromtimestamp(0)
    else:
        time_zone = pytz.UTC
        # epoch = datetime.utcfromtimestamp(0).replace(tzinfo=pytz.UTC)
        epoch = time_zone.localize(datetime.utcfromtimestamp(0))
    return (dt - epoch).total_seconds() * 1000.0


def convert_string_to_milliseconds(ts: str, timezoned: str = None) -> int:
    dt = convert_string_to_datetime(ts=ts, timezoned=timezoned)
    return int(convert_datetime_to_milliseconds(dt=dt, timezoned=timezoned))


def datetime_utc_to_milliseconds(dt: datetime):
    dt = dt.replace(tzinfo=pytz.UTC)
    epoch = datetime.utcfromtimestamp(0).replace(tzinfo=pytz.UTC)
    return (dt - epoch).total_seconds() * 1000.0


def end_time_from_start_time(startTime: int, limit=1000, tick_interval='1h') -> int:
    return startTime + (limit * tick_seconds[tick_interval] * 1000)


def start_time_from_end_time(endTime: int, limit=1000, tick_interval='1h') -> int:
    return endTime - (limit * tick_seconds[tick_interval] * 1000)


def convert_milliseconds_to_local_timezone_string(ms: int) -> str:
    seconds = int(ms) / 1000
    return str(datetime.fromtimestamp(seconds).strftime('%Y-%m-%d %H:%M:%S.%f'))


def convert_milliseconds_to_utc_string(ms: int) -> str:
    seconds = int(ms) / 1000
    return str(datetime.utcfromtimestamp(seconds).strftime('%Y-%m-%d %H:%M:%S.%f'))


def convert_milliseconds_to_utc_datetime(ms: int) -> datetime:
    seconds = int(ms) / 1000
    return datetime.utcfromtimestamp(seconds)


def convert_milliseconds_to_time_zone_datetime(ms: int, timezoned: str = None) -> datetime:
    seconds = int(ms) / 1000
    dt = datetime.utcfromtimestamp(seconds)
    utc_tz = pytz.UTC
    utc_datetime = utc_tz.localize(dt)
    time_zone = pytz.timezone(timezoned)
    return utc_datetime.astimezone(time_zone)


def convert_milliseconds_to_str(ms: int, timezoned: str = None) -> str:
    dt = convert_milliseconds_to_time_zone_datetime(ms=ms, timezoned=timezoned)
    return convert_datetime_to_string(dt)


def utc_start_of_day_ms() -> int:
    # now = get_server_time()
    now = int(time()*1000)
    dt = convert_milliseconds_to_utc_datetime(now).replace(hour=0, minute=0, second=0, microsecond=0)
    return int(convert_datetime_to_milliseconds(dt))


def utc_ms_day_begin_end(ts: int) -> (int, int):
    """Calcula en horario utc el timestamp del comienzo del dia y su final del timestamp pasado como argumento."""
    day_ini_df = convert_milliseconds_to_utc_datetime(ts).replace(hour=0, minute=0, second=0, microsecond=0)
    day_ini_ms = convert_utc_datetime_to_milliseconds(day_ini_df)
    day_end_ms = day_ini_ms + (24 * 60 * 60 * 1000) - 1
    return int(day_ini_ms), int(day_end_ms)


def split_time_interval_in_full_days(ts_ini: int, ts_end: int) -> list:
    """Dado un intervalo mediante dos timestamps, devuelve un alista de tuplas con el inicio y final de cada día completo
     tocado por el intervalo input."""
    ini_first_day, end_first_day_ = utc_ms_day_begin_end(ts_ini)
    ini_last_day_, end_last_day = utc_ms_day_begin_end(ts_end)
    day_ms = 24 * 60 * 60 * 1000
    days = int(ceil_division((end_last_day - ini_first_day), day_ms))
    ret = []
    for i in range(days):
        day_ts_ini = ini_first_day + i * day_ms
        day_ts_end = day_ts_ini + day_ms - 1
        ret.append((int(day_ts_ini), int(day_ts_end)))
    return ret


def time_interval(tick_interval: str,
                  limit: int = 1000,
                  start_time: int = None,
                  end_time: int = None) -> Tuple[int, int]:
    """
    Obtain a timestamp based on ticks intervals from a start or an end timestamp, based on limit.

    If no start or end timestamp passed, then use current utc timestamp in milliseconds and limit.

    :param str tick_interval: A binance valid tick interval
    :param int start_time: A timestamp in milliseconds.
    :param int end_time: A timestamp in milliseconds.
    :param int limit: Ticks limit. Not applied if start and end passed. Default is 1000
    :return: A tuple with the start and end timestamps.

    """
    total_interval_ms = int(tick_seconds[tick_interval] * 1000 * limit)
    if not start_time and not end_time:
        # end_time = utc()
        end_time = int(time()*1000)
        start_time = end_time - total_interval_ms
    elif not end_time and start_time:
        end_time = int(start_time) + total_interval_ms
    elif not start_time and end_time:
        start_time = int(end_time) - total_interval_ms
    return start_time, end_time


def ceil_division(a: float, b: float) -> int:
    """
    Better not to use //.

    From: https://stackoverflow.com/questions/14822184/is-there-a-ceiling-equivalent-of-operator-in-python

    :param float a: A value.
    :param float b: Other value.
    :return int: Returns integer from ceil division.
    """
    return int(-(a // -b))


def ticks_between_timestamps(start: int,
                             end: int,
                             tick_interval: str) -> int:
    """
    Dados dos timestamps en milliseconds y un tick interval te dice cuantas velas entran ahi.

    Se aplica:
        - si el start es un open se toma la vela en la que abre, si no, se toma la siguiente vela.
        - si el end es un open se toma la vela de ese open también, si no, igualmente, la vela que contiene al end.

    El limit no se tiene en cuenta aquí, previamente se habrá tenido en cuenta.
    """
    start_open = open_from_milliseconds(ms=start, tick_interval=tick_interval)
    if start_open != start:
        start_open = next_open_by_milliseconds(ms=start, tick_interval=tick_interval)
    end_close = close_from_milliseconds(end, tick_interval=tick_interval)
    ret = ceil_division(end_close - start_open, tick_seconds[tick_interval]*1000)
    return ret


def get_ticks_to_timestamp_utc(ticks_interval: str, timestamp: int = None):
    """Retorna cuantas velas se requieren hasta el timestamp pasado hasta ahora"""
    # now = get_server_time()
    now = int(time()*1000)
    tick_ms = tick_seconds[ticks_interval] * 1000
    interval_ms = now - timestamp
    return interval_ms // tick_ms


def convert_string_to_datetime(ts: str, timezoned: str = None):
    try:
        ret = datetime.strptime(ts, '%Y-%m-%d %H:%M:%S.%f')
    except ValueError:
        ret = datetime.strptime(ts, '%Y-%m-%d %H:%M:%S')
    if timezoned:
        mytz = pytz.timezone(timezoned)
        return mytz.localize(ret)
    else:
        return ret


def convert_utc_datetime_to_milliseconds(dt):
    epoch = datetime.utcfromtimestamp(0)
    return (dt - epoch).total_seconds() * 1000.0


def utc() -> int:
    """
    Returns UTC time in milliseconds from epoch.
    :return: int
    """
    return int(datetime.now().timestamp() * 1000)


def utc_datetime() -> datetime:
    return datetime.now(pytz.utc)


# Checkers and waiters

def check_tick_interval(tick_interval: str) -> str:
    """
    Checks if argument is a Binance valid tick interval for candles.

    :param str tick_interval: A string, maybe, binance tick interval well formatted.

    """
    if not (tick_interval.lower() in tick_interval_values):
        if not (tick_interval.upper() in tick_interval_values):
            raise Exception(f"BinPan Error on tick_interval: {tick_interval} not in "
                            f"expected API intervals.\n{tick_interval_values}")
        else:
            return tick_interval.upper()
    else:
        return tick_interval.lower()


def detect_tick_interval(data: pd.DataFrame) -> str:
    """Retorna el tick interval de un dataframe"""
    ts_a = data.iloc[0]['Open timestamp']
    ts_b = data.iloc[1]['Open timestamp']
    seconds = (ts_b - ts_a) // 1000
    return list(tick_seconds.keys())[list(tick_seconds.values()).index(seconds)]


def next_open_utc(tick_interval: str) -> int:
    """Calcula el timestamp del next open para un tick_interval"""
    # utc_ms = get_server_time()
    utc_ms = int(time()*1000)
    units = (utc_ms // (tick_seconds[tick_interval] * 1000))
    last_open_ms = units * tick_seconds[tick_interval] * 1000
    return last_open_ms + (tick_seconds[tick_interval] * 1000)


def next_open_by_milliseconds(ms: int, tick_interval: str) -> int:
    """Calcula el timestamp del next open para un tick_interval."""
    units = (int(ms) // (tick_seconds[tick_interval] * 1000))
    last_open_ms = units * tick_seconds[tick_interval] * 1000
    return last_open_ms + (tick_seconds[tick_interval] * 1000)


def open_from_milliseconds(ms: int, tick_interval: str) -> int:
    """Retorna el open de la vela definida que contiene ese timestamp.
    Si coincide el timestamp con un open, calcula y devuelve ese mismo timestamp."""
    units = (int(ms) // (tick_seconds[tick_interval] * 1000))
    return units * tick_seconds[tick_interval] * 1000


def close_from_milliseconds(ms: int, tick_interval: str) -> int:
    """Retorna el close de la vela definida que contiene ese timestamp.
    Si coincide el timestamp con un open, calcula el final."""
    units = (int(ms) // (tick_seconds[tick_interval] * 1000))
    units += 1
    return (units * tick_seconds[tick_interval] * 1000) - 1000


def wait_seconds_until_next_minute():
    # server_time = binpan_modules.api_control.api.get_server_time()
    server_time = int(time() * 1000)  # ahorro de llamadas al server, mejorada sincro de tiempo en windows
    now = datetime.fromtimestamp(server_time / 1000.0).replace(tzinfo=pytz.UTC)
    return 60 - now.second



"""

Tagging utils.

"""
import pandas as pd
import numpy as np
from typing import Tuple

from .exchange import get_info_dic, get_bases_dic, get_quotes_dic
from .market import get_candles_by_time_stamps, parse_candles_to_dataframe
import handlers.logs

logger = handlers.logs.Logs(filename='./logs/tags.log', name='tags', info_level='INFO')


def tag_value(serie: pd.Series,
              value: int or float,
              gt: bool = False,
              ge: bool = False,
              eq: bool = False,
              le: bool = False,
              lt: bool = False,
              match_tag=1,
              mismatch_tag=0,
              to_numeric=True
              ) -> pd.Series:
    """
    It tags values of a serie compared to one value and a gt,ge,eq,le,lt condition.

    :param pd.Series serie: A numeric serie.
    :param int or float or Decimal value: Value of comparison.
    :param bool gt: Value to match if strictly greater than value.
    :param bool ge: Value to match if greater or equal than value.
    :param bool eq: Value to match if equal than value.
    :param bool le: Value to match if lower or equal than value.
    :param bool lt: Value to match if strictly lower than value.
    :param int or str match_tag: Value to tag previous logic.
    :param mismatch_tag: Value to tag not matched previous logic.
    :param bool to_numeric: If true, if possible, downcast type until the most basic numeric type (integer)
    :return pd.Series: A serie with tags as values.
    """
    ret = pd.Series(index=serie.index, dtype=type(match_tag))

    if gt:
        ret = serie.gt(value)
    if ge:
        ret = serie.ge(value)
    if eq:
        ret = serie.eq(value)
    if le:
        ret = serie.le(value)
    if lt:
        ret = serie.lt(value)

    ret.loc[ret] = match_tag
    ret.loc[ret == False] = mismatch_tag
    if to_numeric:
        return pd.to_numeric(arg=ret, downcast='integer')
    else:
        return ret


def tag_comparison(serie_a: pd.Series,
                   serie_b: pd.Series,
                   gt: bool = False,
                   ge: bool = False,
                   eq: bool = False,
                   le: bool = False,
                   lt: bool = False,
                   match_tag=1,
                   mismatch_tag=0,
                   to_numeric=True
                   ) -> pd.Series:
    """
    It tags values of a serie compared to other serie by methods gt,ge,eq,le,lt condition.

    :param pd.Series serie_a: A numeric serie.
    :param pd.Series serie_b: A numeric serie.
    :param bool gt: Value to match if strictly greater than serie_b values.
    :param bool ge: Value to match if greater or equal than serie_b values.
    :param bool eq: Value to match if equal than serie_b values.
    :param bool le: Value to match if lower or equal than serie_b values.
    :param bool lt: Value to match if strictly lower than serie_b values.
    :param int or str match_tag: Value to tag previous logic.
    :param mismatch_tag: Value to tag not matched previous logic.
    :param bool to_numeric: If true, if possible, downcast type until the most basic numeric type (integer)
    :return pd.Series: A serie with tags (argument) as values.
    """
    ret = pd.Series(index=serie_a.index, dtype=type(match_tag))

    if gt:
        ret = serie_a.gt(serie_b)
    if ge:
        ret = serie_a.ge(serie_b)
    if eq:
        ret = serie_a.eq(serie_b)
    if le:
        ret = serie_a.le(serie_b)
    if lt:
        ret = serie_a.lt(serie_b)

    ret.loc[ret] = match_tag
    ret.loc[ret == False] = mismatch_tag

    if to_numeric:
        return pd.to_numeric(arg=ret, downcast='integer')
    else:
        return ret


def tag_cross(serie_a: pd.Series,
              serie_b: pd.Series,
              echo: int = 0,
              cross_over_tag='buy',
              cross_below_tag='sell',
              name='Cross',
              non_zeros: bool = True) -> pd.Series:
    """
    It tags points where serie_a crosses over serie_b or optionally below.

    :param pd.Series serie_a: A numeric serie.
    :param pd.Series serie_b: A numeric serie.
    :param int echo: It tags a fixed amount of candles forward the crossed point not including cross candle.
    :param int or str cross_over_tag: Value to tag over cross. Default is "buy".
    :param int or str cross_below_tag: Value to tag below cross. Default is "sell"
    :param str name: Name for the resulting serie.
    :param bool non_zeros: Substitutes zeros by nans. If echo want to be used, must be used non_zeros.
    :return pd.Series: A serie with tags as values.
    """
    dif = tag_comparison(serie_a=serie_a, serie_b=serie_b, gt=True)
    ret = dif.diff()

    if echo:
        ret = ret.ffill(limit=echo)

    ret.name = name
    if non_zeros:
        ret = ret.replace({'1': cross_over_tag,
                           1: cross_over_tag,
                           '-1': cross_below_tag,
                           -1: cross_below_tag,
                           '0': np.nan,
                           0: np.nan
                           })
    else:
        ret = ret.replace({'1': cross_over_tag, 1: cross_over_tag, '-1': cross_below_tag, -1: cross_below_tag})

    if echo:
        ret = ret.ffill(limit=echo)
    return ret


def tag_column_to_strategy_group(column: str,
                                 group: str,
                                 strategy_groups: dict) -> dict:
    """
    Tags a columns as part of Strategy group of columns.

    :param str column: A column to tag with a strategy group.
    :param str group: Name of the group.
    :param str strategy_groups: The existing strategy groups.
    :return dict: Updated strategy groups of columns.

    """

    if not group in strategy_groups.keys():
        strategy_groups.update({group: [column]})
    else:
        group_columns = strategy_groups[group]
        group_columns.append(column)
        strategy_groups.update({group: group_columns})

    logger.info(f"Added to strategy group {group}: {column}")

    return strategy_groups


################
# merge series #
################


def merge_series(predominant: pd.Series,
                 other: pd.Series) -> pd.Series:
    """
    Predominant serie will be filled nans with values, if existing, from the other serie.
    
    Same kind of index needed.
    
    :param pd.Series predominant: A serie with nans to fill from other serie.
    :param pd.Series other: A serie to pick values for the nans.
    :return pd.Series: A merged serie. 
    """
    return predominant.combine_first(other=other)


def clean_in_out(serie: pd.Series,
                 in_tag=1,
                 out_tag=-1) -> pd.Series:
    """
    Balance from in tag to first out tag and discards any other tag until next in tag. This cleans redundant in or out tags.

    :param pd.Series serie: A serie with in and out tags.
    :param in_tag: Tag for in tags. Default is 1.
    :param out_tag: Tag for out tags. Default is -1.
    :return pd.Series: Clean serie with ech in with next out.
    """

    ret = pd.Series(index=serie.index)

    last_tag = np.nan

    for idx, value in serie.iteritems():

        if value == out_tag and last_tag != out_tag:
            ret[idx] = value
            last_tag = value

        elif value == in_tag and last_tag != in_tag:
            ret[idx] = value

        if value == in_tag or value == out_tag:
            last_tag = value

    return ret


################
# BACK TESTING #
################

def buy_base_backtesting(row: pd.Series, price: float, base: float, quote: float, fee=0.001) -> tuple:
    """
    A simple backtesting buy function.

    :param pd.Series row: A row of a BinPan DataFrame with prices.
    :param float price: Price of the action.
    :param float base: Inverted base.
    :param float quote: Inverted quote.
    :param float fee: Fee applied to the operation.
    :return tuple: A tuple with resulting base and quote after operation.
    """
    high_value = row['High']
    low_value = row['Low']

    if low_value <= price <= high_value:
        ordered_price = price

    elif price > high_value:
        ordered_price = high_value

    else:
        return base, quote

    base += (quote / ordered_price) * (1 - fee)
    quote = 0
    return base, quote


def sell_base_backtesting(row: pd.Series,
                          price: float,
                          base: float,
                          quote: float,
                          fee=0.001) -> tuple:
    """
    A simple backtesting sell function.

    :param pd.Series row: A row of a BinPan DataFrame with prices.
    :param float price: Price of the action.
    :param float base: Inverted base.
    :param float quote: Inverted quote.
    :param float fee: Fee applied to the operation.
    :return tuple: A tuple with resulting base and quote after operation.
    """
    high_value = row['High']
    low_value = row['Low']

    if low_value <= price <= high_value:
        ordered_price = price

    elif price < low_value:
        ordered_price = low_value

    else:
        return base, quote

    quote += base * ordered_price * (1 - fee)
    base = 0
    return base, quote


def evaluate_wallets(df_: pd.DataFrame,
                     base_serie: pd.Series,
                     quote_serie: pd.Series,
                     evaluating_quote: str,
                     info_dic: dict = None,
                     suffix: str = ''
                     ) -> Tuple[pd.Series, pd.Series, pd.Series]:
    """
    Obtains total wallet values in a quote value through time.

    :param pd.DataFrame df_: A BinPan's dataframe.
    :param pd.Series base_serie: A wallet serie for symbol's base.
    :param pd.Series quote_serie: A wallet serie for symbol's quote.
    :param str evaluating_quote: a Binance valid quote to evaluate operations.
    :param dict info_dic: BinPan exchange info dict to extract information about quotes and bases of symbols.
    :param str suffix: A suffix for the names of the columns.
    :return pd.DataFrame: It returns a dataframe with base and quote series plus an evaluated serie in any quote coin of both wallets merged.
    """
    if evaluating_quote:
        evaluating_quote = evaluating_quote.upper()

    original_index = df_.index

    symbol = original_index.name.split()[0]
    tick_interval = original_index.name.split()[1]
    df_.set_index('Open timestamp', inplace=True, drop=False)

    if not info_dic:
        info_dic = get_info_dic()

    bases = get_bases_dic(info_dic=info_dic)
    quotes = get_quotes_dic(info_dic=info_dic)

    base = bases[symbol]
    quote = quotes[symbol]

    if not evaluating_quote:
        evaluating_quote = quote

    if quote == evaluating_quote:
        evaluated_base_serie = df_['Close']
        evaluated_quote_serie = pd.Series(1, index=df_.index)
    else:
        evaluate_base_symbol = base + evaluating_quote
        evaluate_quote_symbol = quote + evaluating_quote

        start_timestamp = int(df_['Open timestamp'].dropna().iloc[0])
        end_timestamp = int(df_['Close timestamp'].dropna().iloc[-1])

        evaluated_base_list = get_candles_by_time_stamps(symbol=evaluate_base_symbol,
                                                         tick_interval=tick_interval,
                                                         start_time=start_timestamp,
                                                         end_time=end_timestamp)
        evaluated_base_df = parse_candles_to_dataframe(raw_response=evaluated_base_list,
                                                       symbol=evaluate_base_symbol,
                                                       tick_interval=tick_interval)
        evaluated_base_df.set_index('Open timestamp', inplace=True, drop=False)

        evaluated_quote_list = get_candles_by_time_stamps(symbol=evaluate_quote_symbol,
                                                          tick_interval=tick_interval,
                                                          start_time=start_timestamp,
                                                          end_time=end_timestamp)
        evaluated_quote_df = parse_candles_to_dataframe(raw_response=evaluated_quote_list,
                                                        symbol=evaluate_quote_symbol,
                                                        tick_interval=tick_interval)

        evaluated_quote_df.set_index('Open timestamp', inplace=True, drop=False)
        # try:
        #     assert df_.index == evaluated_base_df.index
        #     assert df_.index == evaluated_quote_df.index
        # except AssertionError:
        #     raise Exception(f"Index not matching original object index when evaluating quote price.")

        evaluated_base_serie = evaluated_base_df['Close']
        evaluated_quote_serie = evaluated_quote_df['Close']

    # apply qty for the price
    base_value = pd.Series(data=(evaluated_base_serie.values * base_serie.values), index=original_index)
    quote_value = pd.Series(data=(evaluated_quote_serie.values * quote_serie.values), index=original_index)

    # merge data
    merged = base_value + quote_value

    if suffix:
        merged.name = f"Evaluated_{symbol}_in_{evaluating_quote}{suffix}"
    else:
        merged.name = f"Evaluated_{symbol}_in_{evaluating_quote}"
    # merged.index = original_index
    return base_serie, quote_serie, merged


def check_action_labels_for_backtesting(actions: pd.Series,
                                        label_in: int,
                                        label_out: int
                                        ) -> Tuple[any, any]:
    """
    This function verify labels in actions series, if labels for in actions and labels for out actions are or not the specified ones. Also
     if there are any other label it will throw an exception.

    :param pd.Series actions: A serie with labels to check if there are just in and out labels or not.
    :param int label_in: Usually 1
    :param int label_out: Usually -1
    :return tuple: Returns verified label in and label out.
    """
    actions_no_zeros = actions[actions != 0]
    actions_labels = list(actions_no_zeros.dropna().value_counts().index)

    try:
        assert set(actions_labels).issubset({label_in, label_out})
    except AssertionError:
        try:
            assert set(actions_labels).issubset({1, -1})
            label_in = 1
            label_out = -1
        except AssertionError:
            raise Exception(f"BinPan Exception: Backtesting expected labels were not correctly specified: {actions_labels}")
    return label_in, label_out


def simple_backtesting(df: pd.DataFrame,
                       actions: pd.Series or str,
                       base: float = 0,
                       quote: float = 1000,
                       priced_actions_col: str = 'Open',
                       fee: float = 0.001,
                       label_in=1,
                       label_out=-1,
                       suffix: str = '',
                       evaluating_quote: str = 'BUSD',
                       info_dic: dict = None) -> pd.DataFrame:
    """
    Returns two pandas series as base wallet and quote wallet over time. Its expected a serie with strings like 'buy' or 'sell'.
    That serie is called the "actions". If it is available a column with the exact price of the actions, can be passed in
    actions_col parameter by column name or a pandas series, if not, Closed price column can be a good approximation.

    All actions will be considered to buy all base as possible or to sell all base as posible.

    :param pd.DataFrame df: A BinPan dataframe.
    :param pd.Series actions: A pandas series with buy or sell strings for simulate actions.
    :param float base: A starting quantity of symbol's base.
    :param float quote: A starting quantity of symbol's quote.
    :param pd.Series or str priced_actions_col: Column with the prices for the action emulation.
    :param float fee: Binance applicable fee for trading. DEfault is 0.001.
    :param str or int label_in: A label consider as trade in trigger.
    :param str or int label_out: A label consider as trade out trigger.
    :param str evaluating_quote: a Binance valid quote to evaluate operations.
    :param dict info_dic: BinPan exchange info dict to extract information about quotes and bases of symbols.
    :param str suffix: A suffix for the names of the columns.
    :return tuple: Two series with the base wallet and quote wallet funds in time.
    """

    df_ = df.copy(deep=True)

    if type(actions) == str:
        actions_ = df_[actions].copy(deep=True)
    else:
        actions_ = actions.copy(deep=True)

    # check action labels
    label_in, label_out = check_action_labels_for_backtesting(actions=actions_,
                                                              label_in=label_in,
                                                              label_out=label_out)

    base_wallet, quote_wallet = [], []

    last_action = 2314213  # any random thing

    for idx, row in df_.iterrows():

        curr_action = actions_[idx]
        price = df_.loc[idx, priced_actions_col]

        if last_action == label_in:
            base, quote = buy_base_backtesting(row=row, price=price, base=base, quote=quote, fee=fee)
        elif last_action == label_out:
            base, quote = sell_base_backtesting(row=row, price=price, base=base, quote=quote, fee=fee)

        base_wallet.append(base)
        quote_wallet.append(quote)

        last_action = curr_action

    base_serie = pd.Series(base_wallet, index=df_.index, name=f"Wallet_base{suffix}")
    quote_serie = pd.Series(quote_wallet, index=df_.index, name=f"Wallet_quote{suffix}")

    if evaluating_quote:
        base_serie, quote_serie, merged = evaluate_wallets(df_=df_,
                                                           base_serie=base_serie,
                                                           quote_serie=quote_serie,
                                                           evaluating_quote=evaluating_quote,
                                                           info_dic=info_dic,
                                                           suffix=suffix)
        return pd.DataFrame([base_serie, quote_serie, merged]).T

    return pd.DataFrame([base_serie, quote_serie]).T


########################
# ADVANCED BACKTESTING #
########################


def backtesting(df: pd.DataFrame,
                actions_column: pd.Series or str,
                target_column: str or pd.Series,
                stop_loss_column: str or pd.Series,
                entry_filter_column: str or pd.Series = None,
                priced_actions_col: str or pd.Series = 'Open',
                fixed_target: bool = True,
                fixed_stop_loss: bool = True,
                base: float = 0,
                quote: float = 1000,
                fee: float = 0.001,
                label_in=1,
                label_out=-1,
                suffix: str = '',
                evaluating_quote: str = None,
                info_dic: dict = None) -> pd.DataFrame:
    """
    Returns two pandas series as base wallet and quote wallet over time. Its expected a serie with numbers like 1 for ins and -1 for outs.
    By ins and out I mean buys or sells, shorts and repays, etc.

    If it is available a column with the exact price of the actions, can be passed in actions_col parameter by column name or a pandas
     series, if not, Open price column can be a good approximation because operations start at the next candle of the tagged one.

    All actions will be considered to buy all base as possible or to sell all base as posible.

    It supports just long positions backtesting yet.

    :param pd.DataFrame df: A BinPan dataframe.
    :param pd.Series or str actions_column: A pandas series with buy or sell strings for simulate actions.
    :param pd.Series or str target_column: Column with data for operation target values.
    :param pd.Series or str stop_loss_column: Column with data for operation stop loss values.
    :param pd.Series or str entry_filter_column: A serie or colum with ones or zeros to allow or avoid entries.
    :param pd.Series or str priced_actions_col: Column with the prices for the action emulation.
    :param bool fixed_target: Target for any operation will be calculated and fixed at the beginning of the operation.
    :param bool fixed_stop_loss: Stop loss for any operation will be calculated and fixed at the beginning of the operation.
    :param float base: A starting quantity of symbol's base.
    :param float quote: A starting quantity of symbol's quote.
    :param float fee: Binance applicable fee for trading. DEfault is 0.001.
    :param str or int label_in: A label consider as trade in trigger.
    :param str or int label_out: A label consider as trade out trigger.
    :param str evaluating_quote: a Binance valid quote to evaluate operations.
    :param dict info_dic: BinPan exchange info dict to extract information about quotes and bases of symbols.
    :param str suffix: A suffix for the names of the columns.
    :return tuple: Two series with the base wallet and quote wallet funds in time.
    """

    df_ = df.copy(deep=True)

    if type(actions_column) == str:
        actions_data = df_[actions_column].copy(deep=True)
    elif type(actions_column) == pd.Series:
        actions_data = actions_column.copy(deep=True)
    else:
        raise Exception(f"BinPan Backtesting Error: Incorrect type for actions_column -> {type(actions_column)}")

    if type(stop_loss_column) == str:
        stop_loss_data = df_[stop_loss_column].copy(deep=True)
    elif type(stop_loss_column) == pd.Series:
        stop_loss_data = stop_loss_column.copy(deep=True)
    else:
        stop_loss_data = pd.Series(data=np.nan, index=df_.index)

    if type(target_column) == str:
        target_data = df_[target_column].copy(deep=True)
    elif type(target_column) == pd.Series:
        target_data = target_column.copy(deep=True)
    else:
        target_data = pd.Series(data=np.nan, index=df_.index)

    if type(priced_actions_col) == str:
        priced_actions_data = df_[priced_actions_col].copy(deep=True)
    elif type(priced_actions_col) == pd.Series:
        priced_actions_data = priced_actions_col.copy(deep=True)
    else:
        priced_actions_data = df_['Open']

    if type(entry_filter_column) == str:
        entry_filter_data = df_[entry_filter_column].copy(deep=True)
    elif type(entry_filter_column) == pd.Series:
        entry_filter_data = entry_filter_column.copy(deep=True)
    else:
        entry_filter_data = pd.Series(data=1, index=df_.index)

    # check action labels
    label_in, label_out = check_action_labels_for_backtesting(actions=actions_data,
                                                              label_in=label_in,
                                                              label_out=label_out)
    base_wallet, quote_wallet = [], []

    last_action = 2314213  # any random thing
    target = None
    sl = None
    prev_target = 0
    prev_sl = 0
    state = 'out'
    label_in_out = 2

    resulting_actions = pd.Series(data=np.nan, index=df_.index)
    executed_prices = pd.Series(data=np.nan, index=df_.index)

    for idx, row in df_.iterrows():

        curr_action = actions_data.loc[idx]
        price = priced_actions_data.loc[idx]
        curr_entry_filter = entry_filter_data.loc[idx]

        curr_stop_loss = stop_loss_data.loc[idx]
        curr_target = target_data.loc[idx]

        curr_low = df_['Low'].loc[idx]
        curr_high = df_['High'].loc[idx]

        executed_price = None

        if not fixed_target:
            target = max(curr_target, prev_target)
        if not fixed_stop_loss:
            sl = max(curr_stop_loss, prev_sl)

        # catch actions
        if last_action == label_in:
            if curr_entry_filter == 1 and state == 'out':
                base, quote = buy_base_backtesting(row=row, price=price, base=base, quote=quote, fee=fee)
                resulting_actions.loc[idx] = label_in
                executed_prices.loc[idx] = price

                target = curr_target
                sl = curr_stop_loss
                executed_price = None

                state = 'in'

        elif last_action == label_out and state == 'in':  # label outs are processed at the end with sl and targets
            executed_price = price
            target = None
            sl = None

        # check execution if not any previous label
        if executed_price is None and sl and target and state == 'in':
            if curr_low <= sl:
                executed_price = sl
                target = None
                sl = None
            elif curr_high > target:
                executed_price = target
                target = None
                sl = None

        if executed_price and state == 'in':
            # noinspection PyTypeChecker
            base, quote = sell_base_backtesting(row=row, price=executed_price, base=base, quote=quote, fee=fee)
            executed_prices.loc[idx] = executed_price
            state = 'out'
            if last_action == label_in:
                # fast buy and sell
                resulting_actions.loc[idx] = label_in_out
            else:
                resulting_actions.loc[idx] = label_out

        base_wallet.append(base)
        quote_wallet.append(quote)

        last_action = curr_action
        prev_target = curr_target
        prev_sl = curr_stop_loss
        # prev_idx = idx

    base_serie = pd.Series(base_wallet, index=df_.index, name=f"Wallet_base{suffix}")
    quote_serie = pd.Series(quote_wallet, index=df_.index, name=f"Wallet_quote{suffix}")

    # resulting_actions = clean_in_out(serie=resulting_actions, in_tag=label_in, out_tag=label_out)

    resulting_actions.name = f"Resulting_actions_{actions_data.name}"
    executed_prices.name = f"Executed_prices_{actions_data.name}"

    # if evaluating_quote:
    base_serie, quote_serie, merged = evaluate_wallets(df_=df_,
                                                       base_serie=base_serie,
                                                       quote_serie=quote_serie,
                                                       evaluating_quote=evaluating_quote,
                                                       info_dic=info_dic,
                                                       suffix=suffix)

    return pd.DataFrame([base_serie, quote_serie, merged, resulting_actions, executed_prices]).T

    # return pd.DataFrame([base_serie, quote_serie, resulting_actions, executed_prices]).T


def backtesting_short(df: pd.DataFrame,
                      actions_column: pd.Series or str,
                      target_column: str or pd.Series,
                      stop_loss_column: str or pd.Series,
                      entry_filter_column: str or pd.Series = None,
                      priced_actions_col: str or pd.Series = 'Open',
                      fixed_target: bool = True,
                      fixed_stop_loss: bool = True,
                      base: float = 10,
                      quote: float = 0,
                      fee: float = 0.001,
                      label_in=1,
                      label_out=-1,
                      suffix: str = '',
                      evaluating_quote: str = None,
                      info_dic: dict = None) -> pd.DataFrame:
    """
    Returns two pandas series as base wallet and quote wallet over time. Its expected a serie with numbers like 1 for ins and -1 for outs.
    By ins and out I mean buys or sells, shorts and repays, etc.

    If it is available a column with the exact price of the actions, can be passed in actions_col parameter by column name or a pandas
     series, if not, Open price column can be a good approximation because operations start at the next candle of the tagged one.

    All actions will be considered to buy all base as possible or to sell all base as posible.

    It supports just long positions backtesting yet.

    :param pd.DataFrame df: A BinPan dataframe.
    :param pd.Series or str actions_column: A pandas series with buy or sell strings for simulate actions.
    :param pd.Series or str target_column: Column with data for operation target values.
    :param pd.Series or str stop_loss_column: Column with data for operation stop loss values.
    :param pd.Series or str entry_filter_column: A serie or colum with ones or zeros to allow or avoid entries.
    :param pd.Series or str priced_actions_col: Column with the prices for the action emulation.
    :param bool fixed_target: Target for any operation will be calculated and fixed at the beginning of the operation.
    :param bool fixed_stop_loss: Stop loss for any operation will be calculated and fixed at the beginning of the operation.
    :param float base: A starting quantity of symbol's base.
    :param float quote: A starting quantity of symbol's quote.
    :param float fee: Binance applicable fee for trading. DEfault is 0.001.
    :param str or int label_in: A label consider as trade in trigger.
    :param str or int label_out: A label consider as trade out trigger.
    :param str evaluating_quote: a Binance valid quote to evaluate operations.
    :param dict info_dic: BinPan exchange info dict to extract information about quotes and bases of symbols.
    :param str suffix: A suffix for the names of the columns.
    :return tuple: Two series with the base wallet and quote wallet funds in time.
    """

    df_ = df.copy(deep=True)

    if type(actions_column) == str:
        actions_data = df_[actions_column].copy(deep=True)
    elif type(actions_column) == pd.Series:
        actions_data = actions_column.copy(deep=True)
    else:
        raise Exception(f"BinPan Backtesting Error: Incorrect type for actions_column -> {type(actions_column)}")

    if type(stop_loss_column) == str:
        stop_loss_data = df_[stop_loss_column].copy(deep=True)
    elif type(stop_loss_column) == pd.Series:
        stop_loss_data = stop_loss_column.copy(deep=True)
    else:
        stop_loss_data = pd.Series(data=np.nan, index=df_.index)

    if type(target_column) == str:
        target_data = df_[target_column].copy(deep=True)
    elif type(target_column) == pd.Series:
        target_data = target_column.copy(deep=True)
    else:
        target_data = pd.Series(data=np.nan, index=df_.index)

    if type(priced_actions_col) == str:
        priced_actions_data = df_[priced_actions_col].copy(deep=True)
    elif type(priced_actions_col) == pd.Series:
        priced_actions_data = priced_actions_col.copy(deep=True)
    else:
        priced_actions_data = df_['Open']

    if type(entry_filter_column) == str:
        entry_filter_data = df_[entry_filter_column].copy(deep=True)
    elif type(entry_filter_column) == pd.Series:
        entry_filter_data = entry_filter_column.copy(deep=True)
    else:
        entry_filter_data = pd.Series(data=1, index=df_.index)

    # check action labels
    label_in, label_out = check_action_labels_for_backtesting(actions=actions_data,
                                                              label_in=label_in,
                                                              label_out=label_out)
    base_wallet, quote_wallet = [], []

    last_action = 2314213  # any random thing
    target = None
    sl = None
    prev_target = 0
    prev_sl = 0
    state = 'out'
    label_in_out = 2

    resulting_actions = pd.Series(data=np.nan, index=df_.index)
    executed_prices = pd.Series(data=np.nan, index=df_.index)

    for idx, row in df_.iterrows():

        curr_action = actions_data.loc[idx]
        price = priced_actions_data.loc[idx]
        curr_entry_filter = entry_filter_data.loc[idx]

        curr_stop_loss = stop_loss_data.loc[idx]
        curr_target = target_data.loc[idx]

        curr_low = df_['Low'].loc[idx]
        curr_high = df_['High'].loc[idx]

        executed_price = None

        if not fixed_target:
            target = min(curr_target, prev_target)
        if not fixed_stop_loss:
            sl = min(curr_stop_loss, prev_sl)

        # catch actions
        if last_action == label_in:
            if curr_entry_filter == 1 and state == 'out':
                base, quote = sell_base_backtesting(row=row, price=price, base=base, quote=quote, fee=fee)
                resulting_actions.loc[idx] = label_in
                executed_prices.loc[idx] = price

                target = curr_target
                sl = curr_stop_loss
                executed_price = None
                state = 'in'

        elif last_action == label_out and state == 'in':  # label outs are processed at the end with sl and targets
            executed_price = price
            target = None
            sl = None

        # check execution if not any previous label
        if executed_price is None and sl and target and state == 'in':
            if curr_high >= sl:
                executed_price = sl
                target = None
                sl = None
            elif curr_low < target:
                executed_price = target
                target = None
                sl = None

        if executed_price and state == 'in':
            # noinspection PyTypeChecker
            base, quote = buy_base_backtesting(row=row, price=executed_price, base=base, quote=quote, fee=fee)
            executed_prices.loc[idx] = executed_price
            state = 'out'
            if last_action == label_in:
                # fast buy and sell
                resulting_actions.loc[idx] = label_in_out
            else:
                resulting_actions.loc[idx] = label_out

        base_wallet.append(base)
        quote_wallet.append(quote)

        last_action = curr_action
        prev_target = curr_target
        prev_sl = curr_stop_loss
        # prev_idx = idx

    base_serie = pd.Series(base_wallet, index=df_.index, name=f"Wallet_base{suffix}")
    quote_serie = pd.Series(quote_wallet, index=df_.index, name=f"Wallet_quote{suffix}")

    # resulting_actions = clean_in_out(serie=resulting_actions, in_tag=label_in, out_tag=label_out)

    resulting_actions.name = f"Resulting_actions_{actions_data.name}"
    executed_prices.name = f"Executed_prices_{actions_data.name}"

    # if evaluating_quote:
    base_serie, quote_serie, merged = evaluate_wallets(df_=df_,
                                                       base_serie=base_serie,
                                                       quote_serie=quote_serie,
                                                       evaluating_quote=evaluating_quote,
                                                       info_dic=info_dic,
                                                       suffix=suffix)

    return pd.DataFrame([base_serie, quote_serie, merged, resulting_actions, executed_prices]).T



"""

This is the plotting module.

"""
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.express as px
import plotly.figure_factory as ff
import pandas as pd
import numpy as np
from random import choice
from .logs import Logs
from .exceptions import BinPanException

plot_logger = Logs(filename='./logs/plotting.log', name='plotting', info_level='INFO')

plotly_colors = ["aliceblue", "antiquewhite", "aqua", "aquamarine", "azure", "beige", "bisque", "black",
                 "blanchedalmond", "blue", "blueviolet", "brown", "burlywood", "cadetblue", "chartreuse", "chocolate",
                 "coral", "cornflowerblue", "cornsilk", "crimson", "cyan", "darkblue", "darkcyan", "darkgoldenrod",
                 "darkgray", "darkgrey", "darkgreen", "darkkhaki", "darkmagenta", "darkolivegreen", "darkorange",
                 "darkorchid", "darkred", "darksalmon", "darkseagreen", "darkslateblue", "darkslategray",
                 "darkslategrey", "darkturquoise", "darkviolet", "deeppink", "deepskyblue", "dimgray", "dimgrey",
                 "dodgerblue", "firebrick", "floralwhite", "forestgreen", "fuchsia", "gainsboro", "ghostwhite", "gold",
                 "goldenrod", "gray", "grey", "green", "greenyellow", "honeydew", "hotpink", "indianred", "indigo",
                 "ivory", "khaki", "lavender", "lavenderblush", "lawngreen", "lemonchiffon", "lightblue", "lightcoral",
                 "lightcyan", "lightgoldenrodyellow", "lightgray", "lightgrey", "lightgreen", "lightpink",
                 "lightsalmon", "lightseagreen", "lightskyblue", "lightslategray", "lightslategrey", "lightsteelblue",
                 "lightyellow", "lime", "limegreen", "linen", "magenta", "maroon", "mediumaquamarine", "mediumblue",
                 "mediumorchid", "mediumpurple", "mediumseagreen", "mediumslateblue", "mediumspringgreen",
                 "mediumturquoise", "mediumvioletred", "midnightblue", "mintcream", "mistyrose", "moccasin",
                 "navajowhite", "navy", "oldlace", "olive", "olivedrab", "orange", "orangered", "orchid",
                 "palegoldenrod", "palegreen", "paleturquoise", "palevioletred", "papayawhip", "peachpuff", "peru",
                 "pink", "plum", "powderblue", "purple", "red", "rosybrown", "royalblue", "rebeccapurple",
                 "saddlebrown", "salmon", "sandybrown", "seagreen", "seashell", "sienna", "silver", "skyblue",
                 "slateblue", "slategray", "slategrey", "snow", "springgreen", "steelblue", "tan", "teal", "thistle",
                 "tomato", "turquoise", "violet", "wheat", "white", "whitesmoke", "yellow", "yellowgreen"]


def set_color():
    """
    Choose a random color from plotly colors.
    :return:
    """
    return choice(plotly_colors)


def set_subplots(extra_rows: int, candles_ta_height_ratio: float = 0.8, vertical_spacing: float = 0.2):
    """
    Define rows and columns for subplots in plotting engine.

    :param int extra_rows:
    :param float candles_ta_height_ratio:
    :param float vertical_spacing:
    :return:
    """
    # volume is extra row
    ta_rows_heights = [(1 - candles_ta_height_ratio) / extra_rows for _ in range(extra_rows)]
    rows_heights = [candles_ta_height_ratio] + ta_rows_heights
    rows_heights = [float(i) / sum(rows_heights) for i in rows_heights]

    specs = [[{"secondary_y": False}] for _ in range(extra_rows + 1)]
    rows = 1 + extra_rows

    plot_logger.debug(f"rows_heights: {rows_heights}")
    plot_logger.debug(f"sum(rows_heights): {sum(rows_heights)}")
    plot_logger.debug(f"rows: {rows}")
    plot_logger.debug(f"vertical_spacing: {vertical_spacing}")
    plot_logger.debug(f"specs: {specs}")

    return make_subplots(rows=rows, cols=1, shared_xaxes=True, row_heights=rows_heights,
                         vertical_spacing=vertical_spacing, specs=specs)


def set_candles(df: pd.DataFrame,
                x_labels: list = None) -> tuple:
    """
    Put candles and axis into a tuple.
    :param pd.DataFrame df: Dataframe OHLC type.
    :param list x_labels: Labels to replace in x axis plotting.
    :return:
    """
    candles_plot = go.Candlestick(x=df.index, open=df['Open'], high=df['High'], low=df['Low'], close=df['Close'], name='Candles')
    if x_labels:
        # candles_plot.x = x_labels
        candles_plot.x = np.array(x_labels)
    ax = 1
    return candles_plot, ax


# noinspection PyTypeChecker
def set_volume_series(df: pd.DataFrame,
                      win: int = 21) -> tuple:
    """
    Sets or unsets volume histogram.

    :param df: A binpan's dataframe.
    :param win: Window for volume average line.
    :return: A tuple with several figures.
    """
    # volume
    volume_green = df[df['Open'] <= df['Close']]
    volume_red = df[df['Open'] > df['Close']]

    volume_g = go.Bar(x=volume_green.index, y=volume_green['Volume'], marker_color='rgba(144,194,178,255)',
                      name='Up volume')
    volume_r = go.Bar(x=volume_red.index, y=volume_red['Volume'], marker_color='rgba(242,149,149,255)',
                      name='Down volume')
    vol_ewma = df['Volume'].ewm(span=win, min_periods=0, adjust=False, ignore_na=False).mean()
    # volume_ma = set_ta_scatter(df_, vol_ewma)
    volume_ma = go.Scatter(x=df.index, y=vol_ewma, line=dict(color='black', width=0.5), name=f'Volume EMA {win}')
    return volume_g, volume_r, volume_ma, 3


def set_ta_scatter(df: pd.DataFrame,
                   serie: pd.Series,
                   annotations: list = None,
                   color='blue',
                   name='Indicator',
                   text_position="bottom center"):
    """
    Plot a scatter plot for an indicator.

    :param pd.DataFrame df: A binpan's dataframe just for the index.
    :param pd.Series serie: data to plot.
    :param list annotations: Text for annotations over the plotted data.
    :param str color: A plotly color.
    :param str name: Name for the plot.
    :param str text_position: A position from plotly documented annotation positions.
    :return:
    """
    return go.Scatter(x=df.index,
                      y=serie,
                      line=dict(color=color, width=0.1),
                      name=name,
                      mode="markers+text",
                      text=annotations,
                      textposition=text_position)


def set_ta_line(df_index: pd.DataFrame.index,
                serie: pd.Series,
                color='blue',
                name='Indicator',
                line_width: float = 0.5,
                fill_color: str or bool = None,
                fill_mode: str = 'none',
                yaxis: str = 'y',
                show_legend=True):
    """
    Plot a line plot for an indicator.

    :param pd.DataFrame.index df_index: A binpan's dataframe just for the index.
    :param pd.Series serie: data to plot.
    :param str color: A plotly color.
    :param str name: Name for the plot.
    :param float line_width: Line width size.
    :param str or None fill_color: Color to fill areas in the plot.
    :param str fill_mode: Mode to fill for, like to x axis or to next plot, etc.
    :param str yaxis: Axis name for plotting.
    :param bool show_legend: Shows it or not.
    :return:
    """
    my_locals = {k: v for k, v in locals().items() if k != 'df_index' and k != 'serie'}
    plot_logger.debug(f"set_ta_line: {my_locals}")

    if fill_mode:
        fillcolor = fill_color
    else:
        fillcolor = None

    return go.Scatter(x=df_index,
                      y=serie,
                      line=dict(color=color, width=line_width),
                      name=name,
                      mode='lines',
                      fill=fill_mode,
                      fillcolor=fillcolor,
                      yaxis=yaxis,
                      showlegend=show_legend)


def fill_missing(data_list: list, expected_length: int):
    """
    Fills until expected length iterating with the existing values.

    :param list data_list:
    :param int expected_length:
    :return: Filled list.
    """
    ret = []
    cycle = 0
    for i in range(expected_length):
        try:
            ret.append(data_list[i])
        except KeyError:
            if len(data_list) > 0:
                ret.append(data_list[cycle])
                cycle += 1
            else:
                ret.append(f'added_{str(i).zfill(2)}')
    return ret


# noinspection PyTypeChecker
def set_arrows(annotations: pd.Series,
               name: str = None,
               tag: str = None,
               textposition="top center",
               mode="markers+text",
               marker_symbol="arrow-bar-down",
               marker_color='orange',
               marker_line_color='black',
               marker_line_width=0.5,
               marker_size=12):
    """
    Sets arrows.

    Style info at https://plotly.com/python/marker-style/
    """

    if not tag:
        return go.Scatter(mode=mode, x=annotations.index, y=annotations.values, text=annotations.values,
                          marker_symbol=marker_symbol, textposition=textposition, marker_line_color=marker_line_color,
                          marker_color=marker_color, marker_line_width=marker_line_width, marker_size=marker_size,
                          name=name)
    else:
        return go.Scatter(mode=mode, x=annotations.index, y=annotations.values, text=tag, marker_symbol=marker_symbol,
                          textposition=textposition, marker_line_color=marker_line_color, marker_color=marker_color,
                          marker_line_width=marker_line_width, marker_size=marker_size, name=name)


def add_traces(fig,
               list_of_plots: list,
               rows: list,
               cols: list):
    """
    Put traces into the figure.

    :param fig: Figure to set.
    :param list_of_plots: Plot objects to insert into the figure.
    :param list rows: A list with row positions each plot.
    :param list cols: A list with column positions each plot.
    :return: Set figure.
    """
    for i, p in enumerate(list_of_plots):
        # fig.add_trace(p, row=rows[i], col=cols[i], secondary_y=secondary_y)
        fig.append_trace(p, row=rows[i], col=cols[i])
    return fig


def set_layout_format(fig,
                      axis_q: int,
                      title: str,
                      yaxis_title: str,
                      width: int,
                      height: int,
                      range_slider: bool):
    """
    It updates layout of the set plot.

    :param fig: Figure object.
    :param int axis_q: Y axis count.
    :param str title: A title for the layout.
    :param str yaxis_title: Y axis title.
    :param int width: Layout size.
    :param int height: Layout height.
    :param bool range_slider: enabled or not.
    :return:
    """
    layout_kwargs = dict(title=title,
                         yaxis_title=yaxis_title,
                         autosize=False,
                         width=width,
                         height=height,
                         margin=dict(l=1, r=1, b=20, t=100),
                         xaxis_rangeslider_visible=range_slider,
                         xaxis_showticklabels=True)
    # renaming axis names
    for i in range(axis_q):
        axis_name = 'yaxis' + str(i + 1) * (i > 0)
        layout_kwargs[axis_name] = dict(autorange=True, fixedrange=False)  # los subplots pintan bien los datos aunque se expanda el index

    fig = fig.update_layout(layout_kwargs)
    return fig


def update_names(fig, names: dict):
    """
    Update legend names.

    :param fig: Figure object.
    :param names: dict of names.
    :return: Updated figure.
    """
    # new_names = {'col1': 'hello', 'col2': 'hi'}
    fig.for_each_trace(lambda t: t.update(name=names[t.name],
                                          legendgroup=names[t.name],
                                          hovertemplate=t.hovertemplate.replace(t.name, names[t.name])
                                          ))
    return fig


def deploy_traces(annotations: list,
                  colors: list,
                  markers: list,
                  text_positions: list,
                  mark_names: list,
                  tags: list) -> list:
    """

    :param annotations:
    :param colors:
    :param markers:
    :param text_positions:
    :param mark_names:
    :param tags:
    :return:
    """
    length = len(annotations)
    if not colors:
        colors = fill_missing(['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A', '#19D3F3', '#FF6692', '#B6E880',
                               '#FF97FF', '#FECB52'], expected_length=length)
    if not markers:
        markers = fill_missing(["arrow-bar-down", "arrow-bar-up", "arrow-bar-left", "arrow-bar-right"], expected_length=length)
    if not text_positions:
        text_positions = ["top center" for _ in range(len(annotations))]
    if not mark_names:
        mark_names = [f"Annotation {1}" for _ in range(len(annotations))]

    annotations_traces = []  # lista de series con anotaciones
    if tags:
        for idx, an in enumerate(annotations):
            annotations_traces.append(set_arrows(annotations=an,
                                                 textposition=text_positions[idx],
                                                 mode="markers+text",
                                                 marker_symbol=markers[idx],
                                                 marker_color=colors[idx],
                                                 name=mark_names[idx],
                                                 marker_line_color='black',
                                                 marker_line_width=0.5,
                                                 marker_size=15,
                                                 tag=tags[idx]))
    else:
        for idx, an in enumerate(annotations):
            annotations_traces.append(set_arrows(annotations=an,
                                                 textposition=text_positions[idx],
                                                 mode="markers+text",
                                                 marker_symbol=markers[idx],
                                                 marker_color=colors[idx],
                                                 name=mark_names[idx],
                                                 marker_line_color='black',
                                                 marker_line_width=0.5,
                                                 marker_size=15))

    return annotations_traces


###################
# market plotting #
###################

def candles_ta(data: pd.DataFrame,
               indicators_series: list = None,
               rows_pos: list = [],
               indicator_names: list = [],
               indicators_colors: list = [],
               indicators_color_filled: dict = None,
               indicators_filled_mode: dict = None,
               axis_groups: dict = {},
               plot_splitted_serie_couple: dict = {},
               width: int = 1800,
               height: int = 1000,
               range_slider: bool = False,
               candles_ta_height_ratio: float = 0.5,
               plot_volume: bool or str = True,
               title: str = 'Candlesticks, indicators, and Volume plot',
               yaxis_title: str = 'Symbol Price',
               annotation_values: list = None,
               markers: list = None,
               text_positions: list = None,
               annotation_colors: list = None,
               annotation_legend_names: list = None,
               labels: list = None,
               plot_bgcolor: str = None,
               text_index: bool = False):
    """
    Data needs to be a DataFrame that at least contains the columns: Open Close High Low Volume

    It plots candles and optionally volume, but can plot any list of pandas series with indicators (float values) with same index.

    Indicators will be plotted below the candles in subplots according to a row position number, counting 1 as overlay in the candles
    subplot and the rest in row subplots. Several indicators can be plotted in the same row to overlay between them and compare.

    .. note::

       Beware of zeros or values in a different scale when plotting overlapped over candles, that can break the scale of the graph.

    :param pd.DataFrame data: a DataFrame that at least contains the columns: Open Close High Low Volume
    :param list indicators_series: a list of pandas series with float values as indicators.
    :param list rows_pos: 1 means over the candles. Other numbers mean subsequent subplots under the candles.
    :param list indicator_names: Names to show in the plot. Defaults to series name.
    :param list indicators_colors: Color can be forced to anyone from the plotly colors list.

            https://community.plotly.com/t/plotly-colours-list/11730

    :param list or dict indicators_color_filled: Color can be forced to fill to zero line. Is a list of Nones for each indicator in
        indicator list or a fillcolor. For transparent colors use rgba string code to define color. Example for transparent green
        'rgba(26,150,65,0.5)' or transparent red 'rgba(204,0,0,0.5)'. It can be a dictionary with each indicator column name and fill color.
    :param dict indicators_filled_mode: A dict with filled areas for plotting.
    :param dict axis_groups: A dict with named groups for indicators, useful for plotting filled areas using tonexty fill mode.
    :param dict plot_splitted_serie_couple: A dict with splitted data for multiple colours when filling areas using tonexty.
    :param int width: Plot sizing
    :param int height: Plot sizing
    :param bool range_slider: For the volume plot.
    :param float candles_ta_height_ratio: A ratio between the big candles plot and (if any) the rest of indicator subplots below.
    :param bool or str plot_volume: Optional to plot volume from "Volume" column or pass volume column name.
    :param str title: A title string.
    :param str yaxis_title: A name string.
    :param list annotation_values: A list of pandas series with values to plot marks or annotations overlapped in the candles plot.
    :param list markers: Ordered like the annotations list.
        Example

        .. code-block:: python

           markers = ["arrow-bar-down", "arrow-bar-up", "arrow-bar-left", "arrow-bar-right"]

    :param list text_positions: Ordered like the annotations list.
        Example

        .. code-block:: python

           text_positions = ["top center", "middle left", "top center", "bottom center", "top right", "middle left", "bottom right",
            "bottom left", "top right", "top right"]

    :param list annotation_colors: Ordered like the annotations list.
        Example from default colors

        .. code-block:: python

           annotation_colors = ['cornflowerblue', 'blue', 'lightseagreen', 'green', 'cornflowerblue', 'rosybrown', 'lightseagreen',
            'black', 'orange', 'pink', 'red', 'rosybrown', 'cornflowerblue', 'blue', 'lightseagreen', 'green',
            'cornflowerblue', 'rosybrown', 'lightseagreen', 'black', 'orange', 'pink', 'red', 'rosybrown']

    :param list annotation_legend_names: Ordered like the annotations list of names to show in legend.

    :param list labels: Ordered like the annotations list of tags to plot overlapped. It defaults to price value if omitted.

        Example:
        .. code-block:: python

           labels = ['buy', 'sell']

    Plot example:

        .. code-block:: python

            from binpan import binpan

            ethbtc = binpan.Symbol(symbol='ethbtc', tick_interval='1h')

            ethbtc.macd(fast=12, slow=26, smooth=9)
            print(ethbtc.df)

            binpan.handlers.plotting.candles_ta(data=ethbtc.df,
                                                indicators_series=[ethbtc.df['MACD_12_26_9'],
                                                                   ethbtc.df['MACDh_12_26_9'],
                                                                   ethbtc.df['MACDs_12_26_9']],
                                                indicators_color_filled=[False, 'rgba(26,150,65,0.5)', False],
                                                rows_pos=[2, 2, 2],
                                                indicators_colors=['orange', 'green', 'skyblue'])


    .. image:: images/candles_ta_macd.png
        :width: 1000

    :param plot_bgcolor: Set background color.
    :param bool text_index: If enables, index will be transformed to a text index. It can be useful to plot candles not time correlated like reversal candles.

    """
    if not indicators_color_filled and indicators_series:
        indicators_color_filled = {i.name: None for i in indicators_series}
    elif type(indicators_color_filled) == list:
        indicators_color_filled = {s.name: indicators_color_filled[i] for i, s in enumerate(indicators_series)}
    plot_logger.debug(f"candles_ta indicators_color_filled: {indicators_color_filled}")

    if not indicators_filled_mode and indicators_series:
        indicators_filled_mode = {i.name: None for i in indicators_series}
    elif type(indicators_filled_mode) == list:
        indicators_filled_mode = {s.name: indicators_filled_mode[i] for i, s in enumerate(indicators_series)}
    plot_logger.debug(f"candles_ta indicators_filled_mode: {indicators_filled_mode}")

    # catch data
    df_plot = data.copy(deep=True)
    if type(plot_volume) == str:
        df_plot.rename(columns={plot_volume: 'Volume'}, inplace=True)

    if text_index:
        df_plot.index.name = 'Plot Dates'
        df_plot.reset_index(drop=False, inplace=True)
        x_labels = df_plot['Plot Dates'].tolist()
        x_labels = [str(x).split(' ')[1].split('+')[0] for x in x_labels]
    else:
        x_labels = None

    if not indicators_series:
        indicators_series = []

    if not indicators_colors:
        indicators_colors = [choice(plotly_colors) for _ in range(len(indicators_series))]

    if not indicator_names:
        try:
            indicator_names = [i.name for i in indicators_series]
        except Exception:
            indicator_names = [f'Indicator {i}' for i in range(len(indicators_series))]

    if plot_volume:
        extra_rows = len(set(rows_pos)) + 1
    else:
        extra_rows = len(set(rows_pos))

    fig = set_subplots(extra_rows=extra_rows, candles_ta_height_ratio=candles_ta_height_ratio, vertical_spacing=0.02)

    # limit
    axes = 0
    candles_plot, ax = set_candles(df=df_plot, x_labels=x_labels)
    axes += ax

    # volume
    if plot_volume:
        volume_g, volume_r, volume_ma, ax = set_volume_series(df_plot)
        axes += ax
        rows = [1, 2, 2, 2]
        pre_rows = 4
        rows_pos = [i + 1 if i != 1 else i for i in rows_pos]
        traces = [candles_plot, volume_g, volume_r, volume_ma]
    else:
        traces = [candles_plot]
        rows = [1]
        pre_rows = 1

    rows += [i for i in rows_pos]
    cols = [1 for _ in range(len(rows))]

    # technical analysis indicators
    tas = []

    y_axis_idx = [f"y{i}" for i in rows]

    plot_logger.debug(f"----------------------------------------------------------------------")
    plot_logger.debug(f"indicators_colors: {indicators_colors} len: {len(indicators_colors)}")
    plot_logger.debug(f"indicators_color_filled: {indicators_color_filled}")
    plot_logger.debug(f"indicators_filled_mode: {indicators_filled_mode}")
    plot_logger.debug(f"rows: {rows} len: {len(rows)}")
    plot_logger.debug(f"indicators_series: {len(indicators_series)} len: {len(indicators_series)}")
    plot_logger.debug(f"y_axis_idx: {y_axis_idx} len: {len(y_axis_idx)}")
    plot_logger.debug(f"axis_groups: {axis_groups} len: {len(axis_groups)}")
    plot_logger.debug(f"plot_splitted_serie_couple: {plot_splitted_serie_couple} len: {len(plot_splitted_serie_couple)}")
    plot_logger.debug(f"----------------------------------------------------------------------")

    # first get tas with cloud colors "tonexty"
    pre_cached = 0
    for i, indicator in enumerate(indicators_series):
        plot_logger.debug(f"Loop plotting: indicator.name={indicator.name}")
        pre_i = i + pre_rows + pre_cached

        if indicator.name in indicators_filled_mode.keys():
            my_fill_mode = indicators_filled_mode[indicator.name]
        else:
            my_fill_mode = None

        if indicator.name in indicators_color_filled.keys():
            my_fill_color = indicators_color_filled[indicator.name]
        else:
            my_fill_color = None

        if indicator.name in axis_groups.keys():
            my_axis = axis_groups[indicator.name]
        else:
            my_axis = y_axis_idx[pre_i]

        # my_axis_from_cache_100 = f"y1{my_axis[1:]}"

        if indicator_names[i] in plot_splitted_serie_couple.keys():
            plot_logger.debug(f"indicator splitted: {indicator_names[i]}")
            # serie_up, split_up, serie_down, split_down, color_up, color_down = plot_splitted_serie_couple[indicator_names[i]]
            # plot_logger.debug(f"serie_up, split_up, serie_down, split_down, color_up, color_down = {serie_up, split_up, serie_down, split_down, color_up, color_down}")
            indicator_column_up, indicator_column_down, splitted_dfs, color_up, color_down = plot_splitted_serie_couple[indicator_names[i]]
            plot_logger.debug(f"indicator_column_up, indicator_column_down, splitted_dfs,color_up, color_down = "
                              f"{indicator_column_up, indicator_column_down, splitted_dfs, color_up, color_down}")

            tas.append(set_ta_line(df_index=df_plot.index,  # linea para delimitación
                                   serie=indicator,
                                   color=indicators_colors[i],
                                   name=indicator_names[i],
                                   line_width=1,
                                   fill_mode='none',
                                   fill_color=None,
                                   yaxis=my_axis))

            # cambio de función

            def fill_area(label,
                          up_color='rgba(35, 152, 33, 0.5)',
                          down_color='rgba(245, 63, 39, 0.5)'):
                """Internal function to assign colors to areas."""
                if label >= 1:
                    return up_color
                else:
                    return down_color

            for splitted_df in splitted_dfs:
                tas.append(set_ta_line(df_index=splitted_df.index,
                                       serie=splitted_df[indicator_column_up],
                                       color=indicators_colors[i],
                                       name=indicator_column_up,
                                       line_width=0.01,
                                       fill_mode='none',
                                       fill_color=None,
                                       yaxis=my_axis,
                                       show_legend=False))

                tas.append(set_ta_line(df_index=splitted_df.index,
                                       serie=splitted_df[indicator_column_down],
                                       color=indicators_colors[i],
                                       name=indicator_column_down,
                                       line_width=0.01,
                                       fill_mode='tonexty',
                                       fill_color=fill_area(splitted_df['label'].iloc[0]),
                                       yaxis=my_axis,
                                       show_legend=False))

                rows = rows[:pre_i] + [rows[pre_i], rows[pre_i]] + rows[pre_i:]
                pre_cached += 2
                y_axis_idx = y_axis_idx[:pre_i] + [my_axis, my_axis] + y_axis_idx[pre_i:]

            plot_logger.debug(f"rows_updated_by_split: {rows} len: {len(rows)}")
            plot_logger.debug(f"y_axis_idx_updated_by_split: {y_axis_idx} len: {len(y_axis_idx)}")
        else:
            plot_logger.debug(f"indicator_name: {indicator_names[i]}: row: {rows[pre_i]} axis: {my_axis}")

            tas.append(set_ta_line(df_index=df_plot.index,
                                   serie=indicator,
                                   color=indicators_colors[i],
                                   name=indicator_names[i],
                                   line_width=1,
                                   fill_mode=my_fill_mode,
                                   fill_color=my_fill_color,
                                   yaxis=my_axis))
        axes += 1

    cols += [1 for _ in range(len(tas))]
    traces += tas

    # anotaciones, siempre van en la primera fila, la de las velas, son las flechas etc
    if annotation_values:
        annotations_traces = deploy_traces(annotations=annotation_values, colors=annotation_colors, markers=markers,
                                           text_positions=text_positions, mark_names=annotation_legend_names, tags=labels)
        rows += [1 for _ in range(len(annotation_values))]
        cols += [1 for _ in range(len(annotation_values))]
        traces += annotations_traces

    # use different traces for cloud indicators
    fig = add_traces(fig=fig,
                     list_of_plots=traces,
                     rows=rows,
                     cols=cols)

    fig = set_layout_format(fig=fig,
                            axis_q=axes,
                            title=title,
                            yaxis_title=yaxis_title,
                            width=width,
                            height=height,
                            range_slider=range_slider)

    if plot_bgcolor:
        fig.update_layout(plot_bgcolor=plot_bgcolor)

    fig.show()


def candles_tagged(data: pd.DataFrame,
                   width=1800,
                   height=1000,
                   candles_ta_height_ratio=0.5,
                   plot_volume=True,
                   title: str = 'Candlesticks Strategy Plot',
                   yaxis_title: str = 'Symbol Price',
                   on_candles_indicator: list = [],
                   indicator_series: list = [],
                   indicator_names: list = [],
                   indicator_colors: list = [],
                   fill_control: dict or list = None,
                   indicators_filled_mode: dict or list = None,
                   axis_groups: dict or list = {},
                   plot_splitted_serie_couple: dict or list = {},
                   rows_pos: list = [],
                   plot_bgcolor=None,
                   actions_col: str = None,
                   priced_actions_col: str = 'Close',
                   markers_labels: dict = None,
                   markers: dict = None,
                   marker_colors: dict = None,
                   marker_legend_names: dict = None):
    """

    This is a shortcut from candles_ta. It defaults many inputs to better Jupyter Notebook usage.

    Data needs to be a DataFrame that at least contains the columns: Open Close High Low Volume

    It plots candles and optionally volume, but can plot any list of pandas series with indicators (float values) with same index.

    Indicators will be plotted below the candles in subplots according to a row position number, counting 1 as overlay in the candles
    subplot and the rest in row subplots. Several indicators can be plotted in the same row to overlay between them and compare.

    .. note::

        Beware of zeros or values in a different scale when plotting overlapped over candles, that can break the scale of the graph.

    Plot example:

        .. image:: images/plot_tagged.png
            :width: 1000

    :param pd.DataFrame data: a DataFrame that at least contains the columns: Open Close High Low Volume
    :param int width: Plot sizing
    :param int height: Plot sizing
    :param float candles_ta_height_ratio: A ratio between the big candles plot and (if any) the rest of indicator subplots below.
    :param bool plot_volume: Optional to plot volume.
    :param str title: A title string.
    :param str yaxis_title: A name string.
    :param on_candles_indicator: A list of pandas series with values to plot overlapping candles, not in a subplot. Example: SMA.
    :param list indicator_series: a list of pandas series with float values as indicators. Usually not overlap with candles indicators.
        But to plot in a subplot.
    :param list indicator_names: Names to show in the plot. Defaults to series name.
    :param list indicator_colors: Color can be forced to anyone from the plotly colors list:

         https://community.plotly.com/t/plotly-colours-list/11730

    :param dict or list fill_control: A dictionary with color to fill or False bool for each indicator. Is the color to the zero line for
        the indicator plot. If a list passed, it iterates to assign each item in the list with the same index item in the indicators list.
    :param dict indicators_filled_mode: A dict with filled areas for plotting.
    :param dict axis_groups: A dict with named groups for indicators, useful for plotting filled areas using tonexty fill mode.
    :param dict plot_splitted_serie_couple: A dict with splitted data for multiple colours when filling areas using tonexty.
    :param list rows_pos: 1 means over the candles. Other numbers mean subsequent subplots under the candles.
    :param plot_bgcolor: Set background color.
    :param actions_col: A column name of the column with string tags like buy, sell, etc. This is for plotting annotation marks
        overlapped over candles. It is *mandatory* for managing markers, annotations and legend names of annotations.
    :param priced_actions_col: The name of the column containing value of action to position over candles.
     Used just if actions column passed.

        Example:

            .. code-block:: python

                from binpan import binpan
                from handlers.strategies import random_strategy

                bt = binpan.Symbol(symbol='btcusdt',
                                 tick_interval='15m',
                                 time_zone='Europe/Madrid',
                                 end_time='2021-10-31 03:00:00')

                bt.sma(21)

                df = handlers.strategies.random_strategy(data=bt.df, buys_qty=10, sells_qty=12)

                print(df['actions'].value_counts())

                    -1.0    12
                     1.0    10
                    Name: actions, dtype: int64

                binpan.handlers.plotting.candles_tagged(data=df,
                                                        plot_volume=False,
                                                        on_candles_indicator=[df.SMA_21],
                                                        candles_ta_height_ratio=0.8,
                                                        actions_col='actions',
                                                        markers_labels={-1: 'sell', 1: 'buy'},
                                                        marker_colors={-1: 'red', 1: 'green'})

            .. image:: images/plotting/random_strategy_plot.png
               :width: 1000

    :param dict markers_labels: Annotations of tags to show overlapped. Keys are the shown tags in the plot and values are the
     dataframe values to consider what marker and color used. Position of label defaults to close price. This feature is enabled just
     if actions column passed.

        Example:

           .. code-block:: python

              markers_labels = {1: 'buy', -1: 'sell'}


    :param dict markers: Plotly marker type for each label. Usually, if referenced by number will be a not filled mark and using
      string name will be a color filled one. Used just if actions column passed. Check plotly info: https://plotly.com/python/marker-style/

        .. code-block::

            markers = {'1': "arrow-bar-up", '-1': "arrow-bar-down"}

            # for marker integer codes see:

            plotly_markers = [0, '0', 'circle', 100, '100', 'circle-open', 200, '200',
                    'circle-dot', 300, '300', 'circle-open-dot', 1, '1',
                    'square', 101, '101', 'square-open', 201, '201',
                    'square-dot', 301, '301', 'square-open-dot', 2, '2',
                    'diamond', 102, '102', 'diamond-open', 202, '202',
                    'diamond-dot', 302, '302', 'diamond-open-dot', 3, '3',
                    'cross', 103, '103', 'cross-open', 203, '203',
                    'cross-dot', 303, '303', 'cross-open-dot', 4, '4', 'x',
                    104, '104', 'x-open', 204, '204', 'x-dot', 304, '304',
                    'x-open-dot', 5, '5', 'triangle-up', 105, '105',
                    'triangle-up-open', 205, '205', 'triangle-up-dot', 305,
                    '305', 'triangle-up-open-dot', 6, '6', 'triangle-down',
                    106, '106', 'triangle-down-open', 206, '206',
                    'triangle-down-dot', 306, '306', 'triangle-down-open-dot',
                    7, '7', 'triangle-left', 107, '107', 'triangle-left-open',
                    207, '207', 'triangle-left-dot', 307, '307',
                    'triangle-left-open-dot', 8, '8', 'triangle-right', 108,
                    '108', 'triangle-right-open', 208, '208',
                    'triangle-right-dot', 308, '308',
                    'triangle-right-open-dot', 9, '9', 'triangle-ne', 109,
                    '109', 'triangle-ne-open', 209, '209', 'triangle-ne-dot',
                    309, '309', 'triangle-ne-open-dot', 10, '10',
                    'triangle-se', 110, '110', 'triangle-se-open', 210, '210',
                    'triangle-se-dot', 310, '310', 'triangle-se-open-dot', 11,
                    '11', 'triangle-sw', 111, '111', 'triangle-sw-open', 211,
                    '211', 'triangle-sw-dot', 311, '311',
                    'triangle-sw-open-dot', 12, '12', 'triangle-nw', 112,
                    '112', 'triangle-nw-open', 212, '212', 'triangle-nw-dot',
                    312, '312', 'triangle-nw-open-dot', 13, '13', 'pentagon',
                    113, '113', 'pentagon-open', 213, '213', 'pentagon-dot',
                    313, '313', 'pentagon-open-dot', 14, '14', 'hexagon', 114,
                    '114', 'hexagon-open', 214, '214', 'hexagon-dot', 314,
                    '314', 'hexagon-open-dot', 15, '15', 'hexagon2', 115,
                    '115', 'hexagon2-open', 215, '215', 'hexagon2-dot', 315,
                    '315', 'hexagon2-open-dot', 16, '16', 'octagon', 116,
                    '116', 'octagon-open', 216, '216', 'octagon-dot', 316,
                    '316', 'octagon-open-dot', 17, '17', 'star', 117, '117',
                    'star-open', 217, '217', 'star-dot', 317, '317',
                    'star-open-dot', 18, '18', 'hexagram', 118, '118',
                    'hexagram-open', 218, '218', 'hexagram-dot', 318, '318',
                    'hexagram-open-dot', 19, '19', 'star-triangle-up', 119,
                    '119', 'star-triangle-up-open', 219, '219',
                    'star-triangle-up-dot', 319, '319',
                    'star-triangle-up-open-dot', 20, '20',
                    'star-triangle-down', 120, '120',
                    'star-triangle-down-open', 220, '220',
                    'star-triangle-down-dot', 320, '320',
                    'star-triangle-down-open-dot', 21, '21', 'star-square',
                    121, '121', 'star-square-open', 221, '221',
                    'star-square-dot', 321, '321', 'star-square-open-dot', 22,
                    '22', 'star-diamond', 122, '122', 'star-diamond-open',
                    222, '222', 'star-diamond-dot', 322, '322',
                    'star-diamond-open-dot', 23, '23', 'diamond-tall', 123,
                    '123', 'diamond-tall-open', 223, '223',
                    'diamond-tall-dot', 323, '323', 'diamond-tall-open-dot',
                    24, '24', 'diamond-wide', 124, '124', 'diamond-wide-open',
                    224, '224', 'diamond-wide-dot', 324, '324',
                    'diamond-wide-open-dot', 25, '25', 'hourglass', 125,
                    '125', 'hourglass-open', 26, '26', 'bowtie', 126, '126',
                    'bowtie-open', 27, '27', 'circle-cross', 127, '127',
                    'circle-cross-open', 28, '28', 'circle-x', 128, '128',
                    'circle-x-open', 29, '29', 'square-cross', 129, '129',
                    'square-cross-open', 30, '30', 'square-x', 130, '130',
                    'square-x-open', 31, '31', 'diamond-cross', 131, '131',
                    'diamond-cross-open', 32, '32', 'diamond-x', 132, '132',
                    'diamond-x-open', 33, '33', 'cross-thin', 133, '133',
                    'cross-thin-open', 34, '34', 'x-thin', 134, '134',
                    'x-thin-open', 35, '35', 'asterisk', 135, '135',
                    'asterisk-open', 36, '36', 'hash', 136, '136',
                    'hash-open', 236, '236', 'hash-dot', 336, '336',
                    'hash-open-dot', 37, '37', 'y-up', 137, '137',
                    'y-up-open', 38, '38', 'y-down', 138, '138',
                    'y-down-open', 39, '39', 'y-left', 139, '139',
                    'y-left-open', 40, '40', 'y-right', 140, '140',
                    'y-right-open', 41, '41', 'line-ew', 141, '141',
                    'line-ew-open', 42, '42', 'line-ns', 142, '142',
                    'line-ns-open', 43, '43', 'line-ne', 143, '143',
                    'line-ne-open', 44, '44', 'line-nw', 144, '144',
                    'line-nw-open', 45, '45', 'arrow-up', 145, '145',
                    'arrow-up-open', 46, '46', 'arrow-down', 146, '146',
                    'arrow-down-open', 47, '47', 'arrow-left', 147, '147',
                    'arrow-left-open', 48, '48', 'arrow-right', 148, '148',
                    'arrow-right-open', 49, '49', 'arrow-bar-up', 149, '149',
                    'arrow-bar-up-open', 50, '50', 'arrow-bar-down', 150,
                    '150', 'arrow-bar-down-open', 51, '51', 'arrow-bar-left',
                    151, '151', 'arrow-bar-left-open', 52, '52',
                    'arrow-bar-right', 152, '152', 'arrow-bar-right-open']

    :param dict marker_colors: Colors of the annotations. Used just if actions column passed.
    :param dict marker_legend_names: A dict with the names to print as tags over the annotations. Used just if actions column passed.

    """
    data_ = data.copy(deep=True)
    annotations_values = []

    if type(fill_control) == list:
        fill_control = {s.name: fill_control[i] for i, s in enumerate(indicator_series)}

    if type(indicators_filled_mode) == list:
        indicators_filled_mode = {s.name: indicators_filled_mode[i] for i, s in enumerate(indicator_series)}

    if actions_col:  # this trigger all annotation and markers thing
        actions_data = data_[actions_col].dropna()
        actions = sorted(list(set(actions_data.value_counts().index)))

        if not markers_labels:
            markers_labels = {i: i for i in actions}

        try:
            assert len(actions) == len(markers_labels)
        except AssertionError:
            raise Exception(f"BinPan Plotting Exception: Length missmatch between types of actions and markers_labels -> "
                            f"actions={actions} != markers={markers_labels}")

        if not markers:
            my_markers = ["arrow-bar-down", "arrow-bar-up"]
            markers = {mark: my_markers[idx % 2] for idx, mark in enumerate(actions)}

        if not marker_colors:
            my_marker_colors = ['red', 'green', choice(plotly_colors)]
            marker_colors = {mark: my_marker_colors[idx % 3] for idx, mark in enumerate(actions)}
            # marker_colors = {k: choice(plotly_colors) for k, v in markers_labels.items()}

        if not marker_legend_names:
            marker_legend_names = {k: str(v)[0].upper() + str(v)[1:].lower() for k, v in markers_labels.items()}

        for action in actions:  # lista de dataframes por cada acción
            annotations_values.append(data_[data_[actions_col] == action][priced_actions_col])

        # verify annotations, colors, labels and names
        try:
            assert len(markers_labels) == len(markers)
            assert len(markers_labels) == len(marker_colors)
            assert len(markers_labels) == len(marker_legend_names)

        except Exception as exc:
            raise BinPanException(f"Function candles_tagged: Plotting labels, annotation colors or names not consistent with markers list length -> {exc}",
                                  telegram_send=True)
        labels_locator = list(markers_labels.keys())
    else:
        markers_labels = dict()
        markers = dict()
        marker_colors = dict()
        marker_legend_names = dict()
        labels_locator = []

    # indicator allocating rows
    rows_pos_final = []

    if on_candles_indicator:
        rows_pos_final = [1 for _ in range(len(on_candles_indicator))]

    if indicator_series and not rows_pos:
        rows_pos_final += [i + 2 for i in range(len(indicator_series))]
    else:
        rows_pos_final += rows_pos + [i + 2 for i in range(len(indicator_series) - len(rows_pos))]

    indicator_series = on_candles_indicator + indicator_series

    # indicator names for legend
    if not indicator_names:
        try:
            indicator_names = [i.name for i in indicator_series]
        except Exception:
            indicator_names = []
            for i, ind in enumerate(indicator_series):
                try:
                    indicator_names.append(ind.name)
                except:
                    indicator_names.append(f'Indicator_{i}')

    candles_ta(data_,
               width=width,
               height=height,
               range_slider=False,
               candles_ta_height_ratio=candles_ta_height_ratio,
               plot_volume=plot_volume,
               title=title,
               yaxis_title=yaxis_title,
               annotation_values=annotations_values,
               markers=[markers[k] for k in labels_locator],
               labels=[markers_labels[k] for k in labels_locator],
               annotation_colors=[marker_colors[k] for k in labels_locator],
               annotation_legend_names=[marker_legend_names[k] for k in labels_locator],
               rows_pos=rows_pos_final,
               indicators_series=indicator_series,
               indicator_names=indicator_names,
               indicators_colors=indicator_colors,
               indicators_color_filled=fill_control,
               indicators_filled_mode=indicators_filled_mode,
               axis_groups=axis_groups,
               plot_splitted_serie_couple=plot_splitted_serie_couple,
               plot_bgcolor=plot_bgcolor)


################
# trades plots #
################

def plot_trades(data: pd.DataFrame,
                max_size: int = 60,
                height: int = 1000,
                logarithmic: bool = False,
                overlap_prices: pd.DataFrame = None,
                title: str = None,
                shifted: int = 1,
                **kwargs_update_layout):
    """
    Plots scatter plot from trades quantity and trades sizes. Marks are size scaled to the max size. Marks are semi transparent and colored
    using Maker buyer or Taker buyer discrete colors. Usually red and blue.

    Can let you see where are the big sized trades done and the taker or maker buyer side.

    :param pd.DataFrame data: A BinPans trades dataframe.
    :param int max_size: Size of the marks for the biggest quantity sized trades.
    :param int height: Plot sizing.
    :param bool logarithmic: Y axis in a logarithmic scale.
    :param pd.DataFrame overlap_prices: Data to plot overlapping scatter plot.
    :param str title: Title string.
    :param int shifted: If passed any integer, shifts candles to the right one step, this way can see more naturally trades actions over klines.
    :param kwargs_update_layout: Update layout plotly options.

    Example:
        .. code-block:: python

           from binpan import binpan

           lunc = binpan.Symbol(symbol='luncbusd',
                                tick_interval='5m',
                                limit = 100,
                                time_zone = 'Europe/Madrid',
                                time_index = True,
                                closed = True)
           lunc.get_trades()
           binpan.handlers.plotting.plot_trade_size(data = lunc.trades, logarithmic=True)

        .. image:: images/plot_trades_size_log.png
           :width: 1000

    """
    data['Buyer was maker'].replace({False: 'Taker buyer', True: 'Taker Seller'}, inplace=True)
    fig = px.scatter(x=data.index, y=data['Price'], color=data['Buyer was maker'], size=data['Quantity'],
                     size_max=max_size, log_y=logarithmic)
    if not title:
        title = f"Trades size {data.index.name}"

    if type(overlap_prices) == pd.DataFrame:
        start = data.iloc[0]['Timestamp']
        end = data.iloc[-1]['Timestamp']
        # shift added for more reality viewing trades effect on klines
        if shifted:
            title = f"{title} with High and Low Prices (shifted {shifted} candle to the right)"
            plot_data = overlap_prices[(overlap_prices['Open timestamp'] >= start) & (overlap_prices['Open timestamp'] <= end)].shift(1,
                                                                                                                                      freq='infer')
        else:
            title = f"{title} with High and Low Prices"
            plot_data = overlap_prices[(overlap_prices['Open timestamp'] >= start) & (overlap_prices['Open timestamp'] <= end)]

        fig2 = px.line(plot_data, x=plot_data.index, y="High", log_y=logarithmic)
        fig2.update_traces(line=dict(color='rgba(0, 0, 0, 0.6)', width=0.5))

        fig3 = px.line(plot_data, x=plot_data.index, y="Low", log_y=logarithmic)
        fig3.update_traces(line=dict(color='rgba(0, 0, 0, 0.6)', width=0.5))

        fig = go.Figure(data=fig.data + fig2.data + fig3.data)

    fig.update_layout(
        title=title,
        xaxis_title_text=f'{data.index.name}',
        yaxis_title_text=f'Price',
        height=height,
        **kwargs_update_layout)
    fig.show()


##################
# Analysis plots #
##################

def plot_pie(serie: pd.Series,
             categories: int = 15,
             title=f"Size trade categories",
             logarithmic=False):
    """
    Plots a pie chart from a column. Useful to see size ranges in trades, but can be used in any way.

    :param pd.Series serie: pandas serie with numeric values or strings.
    :param int categories: Category count to divide chart.
    :param str title: String title.
    :param bool logarithmic: If logarithmic is selected as true, the sizes of each interval will be distributed in logarithmic steps from
     the smallest to the largest, that is, the smallest values will be divided into smaller groups that will increase exponentially in size.

    Example:

        .. code-block:: python

           from binpan import binpan

           lunc = binpan.Symbol(symbol='luncbusd',
                                tick_interval='5m',
                                limit = 100,
                                time_zone = 'Europe/Madrid',
                                time_index = True,
                                closed = True)

           lunc.get_trades()

           binpan.handlers.plotting.plot_pie(serie = lunc.trades['Quantity'], logarithmic=True)

        .. image:: images/plot_pie_log.png
           :width: 1000

    """
    ma_original = serie.max()
    mi_original = serie.min()
    integer_size = len(str(ma_original).split('.')[0])
    plot_logger.debug(f"integer_size: {integer_size}, max:{ma_original} min:{mi_original}")

    ma = ma_original / 10 ** (integer_size - 1)
    mi = mi_original / 10 ** (integer_size - 1)

    if logarithmic:
        category_steps = 10. ** np.linspace(mi, ma, categories)
        plot_logger.debug(f"category_steps: {category_steps}")
        category_steps = normalize(max_value=ma, min_value=mi, data=category_steps)
        plot_logger.debug(f"category_steps: {category_steps}")

        spread = [mi_original] + normalize(max_value=ma_original, min_value=mi_original, data=category_steps) + [ma_original]
        plot_logger.debug(f"spread: {spread}")
        plot_logger.debug(f"order: {spread}")
    else:
        step = (ma_original - mi_original) / categories
        spread = np.arange(mi_original, ma_original, step)

    # orders = {serie.name: spread}

    pie = serie.groupby(pd.cut(serie, spread)).count()
    names = [str(i) for i in pie.index]

    fig = px.pie(pie,
                 values=serie.name,
                 names=names,
                 color_discrete_sequence=px.colors.sequential.RdBu,
                 title=title,
                 hover_name=serie.name)
    # category_orders=orders)
    fig.show()


def plot_scatter(df: pd.DataFrame,
                 x_col: str,
                 y_col: str,
                 symbol: str = None,
                 color: str = None,
                 marginal: bool = True,
                 title: str = None,
                 height: int = 1000,
                 **kwargs):
    """
    Plot scatter plots with a column of values in X axis and other in Y axis.

    :param pd.DataFrame df: A Dataframe.
    :param str x_col: Name of column with X axis data.
    :param str y_col: Name of column with Y axis data.
    :param str symbol: Name of column with values (discrete or not) to apply a symbol each.
    :param str color: Name of column with values (discrete or not) to apply a color each.
    :param bool marginal: Lateral auxiliar plots.
    :param str title: A title string.
    :param height: Plot sizing.
    :param kwargs: Optional plotly kwargs.

    Example:

    .. code-block::

       from binpan import binpan

       lunc = binpan.Symbol(symbol='luncbusd',
                            tick_interval='5m',
                            limit = 100,
                            time_zone = 'Europe/Madrid',
                            time_index = True,
                            closed = True)

       binpan.handlers.plotting.plot_scatter(df = lunc.df,
                                             x_col='Close',
                                             y_col='Volume',
                                             color='Trades',
                                             symbol='Close',
                                             title='Scatter plot for LUNCBUSD Close price in X and Volume in Y'
                                             )

    .. image:: images/scatter_example.png
       :width: 1000

    """
    if marginal:
        fig = px.scatter(df,
                         x=x_col,
                         y=y_col,
                         symbol=symbol,
                         color=color,
                         title=title,
                         marginal_x="histogram",
                         marginal_y="rug",
                         height=height,
                         **kwargs)
    else:
        fig = px.scatter(df,
                         x=x_col,
                         y=y_col,
                         symbol=symbol,
                         color=color,
                         title=title,
                         height=height,
                         **kwargs)
    fig.show()


def plot_hists_vs(x0: pd.Series,
                  x1: pd.Series,
                  x0_name: str = None,
                  x1_name: str = None,
                  bins: int = 50,
                  hist_funct: str = 'sum',
                  height: int = 900,
                  title: str = None,
                  **kwargs_update_layout):
    """
    Plots two histograms with same x scale to campare distributions of values.

    :param pd.Series x0: A pandas series.
    :param pd.Series  x1: A pandas series.
    :param str x0_name: Name for the legend
    :param str x1_name: Name for the legend
    :param int bins: Number of bins or bars to show.
    :param str hist_funct: A function to apply to data. It can be 'sum', 'count', 'average', etc...

        More details in: https://plotly.com/python/histograms/#histograms-with-gohistogram

    :param int height: Plot sizing.
    :param str title: Plot title.
    :param kwargs_update_layout: Plotly update layout options.


    Example:

    .. code-block::

       from binpan import binpan

       lunc = binpan.Symbol(symbol='luncbusd',
                            tick_interval='5m',
                            limit = 100,
                            time_zone = 'Europe/Madrid',
                            time_index = True,
                            closed = True)

       binpan.handlers.plotting.plot_hists_vs(x0=lunc.df['High'],
                                              x1=lunc.df['Low'],
                                              bins=50,
                                              hist_funct='count',
                                              title='High and Low prices distribution.')

    .. image:: images/hist_vs_dist.png
       :width: 1000

    """
    if not x0_name:
        x0_name = x0.name
    if not x1_name:
        x1_name = x1.name

    fig = go.Figure()

    start = min(x0.min(), x1.min())
    end = max(x0.max(), x1.max())

    fig.add_trace(go.Histogram(x=x0,
                               histfunc=hist_funct,
                               name=x0_name,
                               xbins=dict(
                                   start=start,
                                   end=end,
                                   size=(x0.max() - x0.min()) / bins)))

    fig.add_trace(go.Histogram(x=x1,
                               histfunc=hist_funct,
                               name=x1_name,
                               xbins=dict(
                                   start=start,
                                   end=end,
                                   size=(x0.max() - x0.min()) / bins)))

    fig.update_layout(
        bargap=0.3,
        title=title,
        xaxis_title_text=f'{x0_name} vs {x1_name} size',
        yaxis_title_text=f'{x0_name} vs {x1_name} {hist_funct}',
        bargroupgap=0.1,
        height=height,
        **kwargs_update_layout)

    fig.update_traces(opacity=0.75)
    fig.show()


def orderbook_depth(df: pd.DataFrame,
                    accumulated=True,
                    title='Depth orderbook plot',
                    height=500,
                    plot_y="Quantity",
                    **kwargs):
    """
    Plots orderbook from a BinPan orderbook dataframe.

    :param pd.DAtaFrame df: BinPan orderbook dataframe.
    :param bool accumulated: If true, applies cumsum to asks and bids.
    :param str title: A title string.
    :param int height: Plot sizing.
    :param str plot_y: Column name with y axis data. Defaults to Quantity.
    :param kwargs: Plotly kwargs.

    Example:

    .. code-block::

        from binpan import binpan

        lunc = binpan.Symbol(symbol='luncbusd',
                            tick_interval='5m',
                            limit = 100,
                            time_zone = 'Europe/Madrid',
                            time_index = True,
                            closed = True)

        lunc.get_orderbook()

    .. image:: images/plot_orderbook.png
       :width: 1000

    """
    ob = df.copy(deep=True)
    if accumulated:
        c_asks = ob[ob['Side'] == 'ask']['Quantity'][::-1].cumsum()
        c_bids = ob[ob['Side'] == 'bid']['Quantity'].cumsum()
        cumulated = pd.concat([c_asks, c_bids])
        ob.loc[:, 'Accumulated Quantity'] = cumulated
        plot_y = 'Accumulated Quantity'

    fig = px.line(ob, x="Price", y=plot_y, color='Side', height=height, title=title, **kwargs)
    fig.show()


def dist_plot(df: pd.DataFrame,
              x_col: str = 'Price',
              color: str = 'Side',
              bins: int = 300,
              histnorm: str = 'density',
              height: int = 800,
              title: str = "Distribution",
              **update_layout_kwargs):
    """
    Plot a distribution plot for a dataframe column. Plots line for kernel distribution.

    :param pd.DataFrame df: A BinPan Dataframe like orderbook, candles, or any other.
    :param str x_col: Column name for x-axis data.
    :param str color: Column name with tags or any values for using as color scale.
    :param int bins: Columns in histogram.
    :param str histnorm: One of 'percent', 'probability', 'density', or 'probability density' from plotly express documentation.
        https://plotly.github.io/plotly.py-docs/generated/plotly.express.histogram.html
    :param int height: Plot sizing.
    :param str title: A title string

    Example from binpan Symbol plot_orderbook_density method.

    .. image:: images/orderbook_density.png
       :width: 1000

    """
    filtered_df = df.copy()

    fig = ff.create_distplot(hist_data=[filtered_df["Price"].tolist()],
                             group_labels=["Price"],
                             show_hist=False,
                             ).add_traces(
        px.histogram(filtered_df, x=x_col, nbins=bins, color=color, histnorm=histnorm)
        .update_traces(yaxis="y3", name=x_col)
        .data)

    fig.update_layout(height=height, title=title, yaxis3={"overlaying": "y", "side": "right"}, showlegend=True, **update_layout_kwargs)
    fig.show()


def bar_plot(df: pd.DataFrame,
             x_col_to_bars: str,
             y_col: str,
             bar_segments: str = None,
             bins: int = 100,
             aggregation: str = 'sum',
             height: int = 800,
             title: str = "Bar Plot",
             y_axis_title: str = None,
             legend_names: dict = None,
             **update_layout_kwargs):
    """
    Plot a bar plot for a dataframe column.

    :param pd.DataFrame df: A BinPan Dataframe like orderbook, candles, trades or any other.
    :param str x_col_to_bars: A column name to group values into x bars, like in example, price.
    :param str y_col: Column names for y-axis data, if more than one will be stacked.
    :param str bar_segments: Other column name to differentiate each bar segment.
    :param int bins: Columns in histogram.
    :param str aggregation: Aggregation method, can be 'sum' or 'mean'.
    :param int height: Plot sizing.
    :param str title: A title string
    :param str y_axis_title: Title for y axis plot.
    :param dict legend_names: A dict to rename legend names.

    """
    _df = df.copy(deep=True)

    slot = group_slot(data=_df, master_column=x_col_to_bars, bins=bins)

    if bar_segments:
        discrete_data = [pd.cut(_df[x_col_to_bars], np.arange(_df[x_col_to_bars].min(), _df[x_col_to_bars].max(), slot)), bar_segments]
    else:
        discrete_data = [pd.cut(_df[x_col_to_bars], np.arange(_df[x_col_to_bars].min(), _df[x_col_to_bars].max(), slot))]

    if aggregation == 'sum':
        grouped_data = _df.groupby(discrete_data).sum()
    elif aggregation == 'mean':
        grouped_data = _df.groupby(discrete_data).mean()
    else:
        msg = f"BinPan Exception: Aggregation method must be in ['sum', 'mean'] -> {aggregation}"
        plot_logger.error(msg)
        raise Exception(msg)

    if bar_segments:
        grouped_data.reset_index(inplace=True, level=1)
        col_name_1 = f"agg_{y_col.replace(' ', '_')}_{bar_segments.replace(' ', '_')}"
        col_name_2 = f"agg_{y_col.replace(' ', '_')}_not_{bar_segments.replace(' ', '_')}"
        grouped_data[col_name_1] = grouped_data.loc[grouped_data[bar_segments]][y_col]
        grouped_data[col_name_2] = grouped_data.loc[~grouped_data[bar_segments]][y_col]
        y_columns = [col_name_1, col_name_2]
    else:
        y_columns = [y_col]

    if bar_segments:
        grouped_data.drop(x_col_to_bars, axis=1, inplace=True)
        grouped_data.reset_index(inplace=True)
        grouped_data.drop_duplicates(x_col_to_bars, keep='first', inplace=True)
    else:
        grouped_data.drop(x_col_to_bars, axis=1, inplace=True)
        grouped_data.reset_index(inplace=True)

    if not y_axis_title:
        if bar_segments:
            y_axis_title = str(y_columns).replace('_', ' ').replace("'", '')
        else:
            y_axis_title = y_col

    grouped_data.loc[:, x_col_to_bars] = grouped_data[x_col_to_bars].astype('str')

    # print(grouped_data)

    fig = px.bar(data_frame=grouped_data,
                 x=x_col_to_bars,
                 y=y_columns,
                 barmode='group')

    fig.update_layout(
        bargap=0.3,
        title=title,
        xaxis_title_text=f'{x_col_to_bars} in {bins} intervals.',
        yaxis_title_text=y_axis_title,
        bargroupgap=0.1,
        height=height,
        showlegend=True,
        **update_layout_kwargs)
    if legend_names:
        # fig.for_each_trace(lambda t: print(t.name))
        fig.for_each_trace(lambda t: t.update(name=legend_names[t.name]))
    fig.show()


##############
# plot tools #
##############

def normalize(max_value: int or float, min_value: int or float, data: list):
    """
    Normalize data from minimum as 0 to maximum as 1.

    :param int or float max_value: A numeric value.
    :param int or float min_value: A numeric value.
    :param data: List of numerica data.
    :return: Normalized numeric data.
    """
    return [(i / sum(data)) * max_value + min_value for i in data]


def find_step_for_bins(data: pd.DataFrame,
                       master_column: str,
                       bins: int = 100,
                       pip_size=0.000001):
    """
    Find precision for getting a target number of bins intervals for values in master column.

    Very expensive function.

    :param pd.DataFrame data: A binpan's dataframe.
    :param str master_column: Column to groupby.
    :param int bins: Expected vertical slots.
    :param pip_size: Size of what will be considered a tiny value.
    :return float: The step for the bins.
    """
    trad_min = data[master_column].min()
    trad_max = data[master_column].max()

    curr_bins = len(data)
    curr_pips = pip_size

    while curr_bins > bins:

        cats, bins_result = pd.cut(data[master_column], np.arange(trad_min, trad_max, curr_pips), retbins=True)
        curr_bins = len(bins_result)

        if curr_bins <= bins:
            break
        else:
            curr_pips += pip_size

    return curr_pips


def group_slot(data: pd.DataFrame,
               master_column: str,
               bins: int = 100):
    """
    Get what vertical slot size results for a number of bins.

    :param pd.DataFrame data: A binpan's dataframe.
    :param str master_column: Column to groupby.
    :param int bins: Expected vertical slots.
    :return: Size of each slot.
    """
    p_max = data[master_column].max()
    p_min = data[master_column].min()
    return (p_max - p_min) / bins



import pandas as pd
from csv import QUOTE_ALL
from time import time
from sys import path
from os import listdir, path, mkdir, replace, makedirs

from .starters import AesCipher
from .logs import Logs

files_logger = Logs(filename='./logs/files_logger.log', name='files_logger', info_level='INFO')

cipher_object = AesCipher()


################
# Files manage #
################

def create_dir(my_path: str):
    """
    Create a folder at the designated path if it doesn't exist. No error if it already exists

    :param str my_path: Path for the new directory.

    """
    if not path.exists(my_path):
        makedirs(my_path)


def save_dataframe_to_csv(filename: str,
                          data: pd.DataFrame,
                          col_sep=',',
                          index=False,
                          timestamp=True) -> None:
    """
    Save a dataframe in a csv with a separator between columns. Each field of each column will be enclosed in double quotes.

    :param str filename: The file name.
    :param pd.DataFrame data: The dataframe with date in columns.
    :param str col_sep: Default is ','
    :param bool index: Keep index or not when saving. DEfault is False (drop index)
    :param bool timestamp: Add a timestamp to the file name automatically. Default is true.

    """
    if filename.lower().endswith(".csv"):
        filename = filename.replace('.csv', '')
    if timestamp:
        filename = filename + '_' + str(time()).split('.')[0]
    filename += '.csv'
    data.to_csv(filename, sep=col_sep, header=True, encoding='utf-8', quoting=QUOTE_ALL, index=index)


def append_dataframe_to_csv(filename: str, data: pd.DataFrame, col_sep: str = ',', index: bool = False, header=True) -> None:
    """
    Add lines to a csv with separator to choose, a dataframe. Each field of each column will be enclosed in quotes

    :param str filename: The file to use.
    :param pd.DataFrame data: Data to append to the file.
    :param str col_sep: Expected column separators. Default is ","
    :param bool index: Keeps index in file as first column/s.
    :param bool header: Keeps header. Default is True.
    """
    if path.isfile(filename):
        header = False
    if filename.lower().endswith(".csv"):
        filename = filename.replace('.csv', '')
    filename += '.csv'
    data.to_csv(filename, sep=col_sep, header=header, encoding='utf-8', quoting=QUOTE_ALL, index=index, mode='a')


def find_csvs_in_path(files_path: str = '.', extension='csv'):
    """
    Locate all files with desired extension.

    :param str files_path: A path.
    :param str extension: A extension like "csv".
    :return:
    """
    ret = []
    for r in listdir(files_path):
        full_path = path.join(files_path, r)
        if path.isfile(full_path):
            ret.append(full_path)
    return [f for f in ret if f.endswith(extension)]


def move_old_csvs(files_path: str = '.', extension='csv'):
    """
    Move all existing files in a path with desired extension to a folder called "old".

    :param str files_path: A path.
    :param str extension: A extension. Default is "csv".

    """
    files_path = path.abspath(files_path)
    old = path.join(files_path, 'old')
    if not path.exists(old):
        mkdir(old)
    csvs = find_csvs_in_path(files_path=files_path, extension=extension)
    for file in csvs:
        file_name = path.basename(file)
        dst = path.join(files_path, 'old', file_name)
        print(f'Moving old file {file}')
        replace(file, dst)


def read_csv_to_dataframe(filename: str,
                          col_sep: str = ',',
                          index_col: bool = None) -> pd.DataFrame:
    """
    Creates a csv file from a dataframe.

    :param str filename: The file name with or without path.
    :param str col_sep: Column separator. DEfault is ","
    :param bool index_col: If False, index is dropped. Default is False.
    :return pd.DataFrame: A dataframe with data in columns using file rows header.
    """
    return pd.read_csv(filepath_or_buffer=filename, sep=col_sep, index_col=index_col, skip_blank_lines=True,
                       quoting=QUOTE_ALL)


def read_file(filename: str) -> list:
    """
    Read a file to a list of strings each line.

    :return list: list with a string each row in the file.

    """
    if not path.isfile(filename):
        return []
    with open(filename, 'r') as f:
        lines = f.readlines()
        lines = [line.rstrip() for line in lines]
    return lines


def save_file(filename: str, data: list, mode='w') -> None:
    """
    Save a new file from a list of lists each line.

    :param str filename: a file name to save.
    :param list data: Data in a list of strings each line.
    :param str mode: 'w' to rewrite full file or 'a' to append to existing file.

    """
    with open(filename, mode) as f:
        for line in data:
            f.write(str(line) + '\n')


def select_file(path='.',
                extension='csv') -> str:
    """
    Selects from files in the path with the extension passed.

    :param str path: Path to search files.
    :param str extension: Extension of interesting files to select.
    :return str: a filename.
    """
    files = find_csvs_in_path(files_path=path, extension=extension)
    files = [i for i in files if i.lower().endswith(extension.lower())]
    for i, file in enumerate(files):
        print(f"{i}: {file}")
    selection = input("Insert file number: ")

    return files[int(selection)]

###################
# API AND SECRETS #
###################


def add_api_key(api_key_value: str) -> None:
    """
    Checks if exists in secret.py file and if not, then adds a line with the API key (not secret API key) encrypted value for working
    with the package.

    :param str api_key_value: API key
    """
    filename = "secret.py"
    saved_data = read_file(filename=filename)
    lines = []
    for line in saved_data:
        if not line.startswith('api_key') and line:
            lines.append(line)
    encryptor = cipher_object.encrypt(api_key_value)
    lines.append(f'\napi_key = "{encryptor}"')
    save_file(filename=filename, data=lines)


def add_api_secret(api_secret_value: str) -> None:
    """
    Checks if exists in secret.py file and if not, then adds a line with the API secret encrypted value for working with the package.

    :param str api_secret_value: API key
    """
    filename = "secret.py"
    saved_data = read_file(filename=filename)
    lines = [line for line in saved_data if (not line.startswith('api_secret')) and line]
    encryptor = cipher_object.encrypt(api_secret_value)
    lines.append(f'\napi_secret = "{encryptor}"')
    save_file(filename=filename, data=lines)


def add_any_key(key: str, key_name: str) -> None:
    """
    Checks if exists in a file and if not, then adds a line with the api key value for working with the package.

    :param str key: Any API key or secret to encrypt in secret.py file.
    :param str key_name: Variable name to import it later.

    Example to add telegram data for notifications handler:

        .. code-block:: python

            from binpan import handlers

            handlers.add_any_key(key="xxxxxxxxxx", key_name="encoded_chat_id")
            handlers.add_any_key(key="xxxxxxxxxx", key_name="encoded_telegram_bot_id")


    """
    filename = "secret.py"
    saved_data = read_file(filename=filename)
    lines = []
    for line in saved_data:
        if line:
            if not line.startswith(key_name):
                lines.append(line)
    encryptor = cipher_object.encrypt(key)
    lines.append(f'\n{key_name} = "{encryptor}"')
    save_file(filename=filename, data=lines)
